/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.control;

import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.datalog.GammaToDatalogRewriter;
import de.unibremen.informatik.tdki.grind.datalog.OmegaToDatalogRewriter;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaSaturator;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.omegaconstruction.Omega;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaLoopChecker;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaSaturator;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaTuple;
import de.unibremen.informatik.tdki.grind.statistics.DetailsLogEntry;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.cq.CQException;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogProgram;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;

/**
 *
 * @author hansen
 */
public class GammaAndOmegaCaller {

//    private static int GAMMA_TIMEOUT = 15 * S;
//    private static int OMEGA_TIMEOUT = 15 * S;

    private static final boolean DEBUG = false;

    private final OWLOntologyManager manager;
    private File file;
    private OWLOntology fromFileOnt;
    private int noOfCNs;
    private Context context;
    private long tStartOfThisConcept;
    private final long t0;
    private Set<OWLClass> conceptNames;
    private PrintWriter output;
    private int numberOfGammaCycles;
    private int numberOfGammaCycleFOno, numberOfGammaCycleFOyes;
    private int numberOfGammaTimeouts;
    private int numberOfOmegaTimeouts;
    private final boolean verbose;

    private DatalogProgram resultingProgram;

    public GammaAndOmegaCaller(String ontFilename) { //, boolean logging, boolean verbose, boolean norewrite) {
        // Start global time count:
        t0 = System.currentTimeMillis();

        this.manager = OWLManager.createOWLOntologyManager();
        this.fromFileOnt = null;
        this.verbose = CQContext.getINSTANCE().isVerbose();
        openOntology(ontFilename);
    }

    public GammaAndOmegaCaller(OWLOntology ont) throws CQException {
        // Start global time count:
        t0 = System.currentTimeMillis();

        this.manager = OWLManager.createOWLOntologyManager();
        this.fromFileOnt = ont;
        this.verbose = CQContext.getINSTANCE().isVerbose();
        openOntology(ont);
    }

    public DatalogProgram getResultingProgram() {
        return resultingProgram;
    }

    /**
     * Grind2 addition, run Gamma and Omega with predefined OWLClass instance.
     *
     * @param startconcept
     * @return
     * @throws Exception
     */
    public int runSingleConceptname(OWLClass startconcept) throws Exception {

        tStartOfThisConcept = System.currentTimeMillis();
        DetailsLogEntry detailsLogEntry = new DetailsLogEntry();
        noOfCNs = 1;
        int retval = runGammaAndOmega(startconcept, 1, detailsLogEntry);
        detailsLogEntry.setTimeSpent(System.currentTimeMillis() - tStartOfThisConcept);
        return retval;
    }

    /**
     *
     * @param startConceptString
     * @return return value for shell: 0=FO-rewr, 1=Not FO-rewr.
     * @throws OWLOntologyCreationException
     * @throws FileNotFoundException
     * @throws Exception
     */
    @Deprecated
    public int runSingleConceptname(String startConceptString)
            throws OWLOntologyCreationException, FileNotFoundException, Exception {

        OWLClass startConcept = null;
        boolean found = false;
        for (OWLClass concept : conceptNames) {
            if (StringTools.getFragment(concept.getIRI()).equals(startConceptString)) {
                startConcept = concept;
                found = true;
                break;
            }
        }
        if (!found) {
            System.err.println("grind: Argument " + startConceptString + " is not a concept name in given ontology. Aborting.");
            System.exit(-1);
        }

        tStartOfThisConcept = System.currentTimeMillis();
        DetailsLogEntry detailsLogEntry = new DetailsLogEntry();
        noOfCNs = 1;
        int retval = runGammaAndOmega(startConcept, 1, detailsLogEntry);
        if (verbose) {
            System.out.println("\nTotal time: " + (System.currentTimeMillis() - tStartOfThisConcept) + " ms.\n");
        }
        detailsLogEntry.setTimeSpent(System.currentTimeMillis() - tStartOfThisConcept);
        return retval;
    }

    private void openOntology(String ontSource) {

        if (verbose) {
            System.out.println("\n\nOpening ontology " + ontSource + "...");
        }
        file = new File(ontSource);
        try {
            fromFileOnt = manager.loadOntologyFromOntologyDocument(file);
        } catch (OWLOntologyCreationException e) {
            System.err.println("Error: Could not load ontology: " + e);
            System.exit(-1);
        }
        context = Context.getINSTANCE();
        context.setOntAndManager(fromFileOnt, manager);

        manager.removeOntology(fromFileOnt);

        if (verbose) {
            System.out.println("  Normalising the ontology: " + (System.currentTimeMillis() - t0) + " ms.");
            System.out.println("\n  Size of sub(T):           " + context.getOriginalOnt().getSubconcepts().size());
        }
        conceptNames = context.getOriginalOnt().getConceptNames();
        noOfCNs = conceptNames.size();
    }

    private void openOntology(OWLOntology ont) {

        context = Context.getINSTANCE();
        context.setOntAndManager(ont, manager);
        manager.removeOntology(ont);

        if (verbose) {
            System.out.println("  Normalising the ontology: " + (System.currentTimeMillis() - t0) + " ms.");
            System.out.println("  Size of sub(T):           " + context.getOriginalOnt().getSubconcepts().size());
        }
        conceptNames = context.getOriginalOnt().getConceptNames();
        noOfCNs = conceptNames.size();
    }

    /**
     * retvals: 0 = FOrewr, 1 = nonFOrewr (Omega cycle), -1 = Gamma timeout, -2
     * = Omega timeout.
     *
     * @param startConcept
     * @param currentClass
     * @param detailsLogEntry
     * @return
     * @throws Exception
     */
    private int runGammaAndOmega(OWLClass startConcept, int currentClass, DetailsLogEntry detailsLogEntry) throws Exception {

        detailsLogEntry.setConceptName(startConcept);

        // * Saturate Gamma:
        GammaSaturator gammaSat = new GammaSaturator(startConcept, detailsLogEntry, context);
        long usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
        gammaSat.saturate(CQContext.getINSTANCE().getTimeout() - usedUntilNow);

        // Main loop, do Gamma - Omega - Gamma - Omega...
        while (true) {

            // * if Gamma timeout:
            if (detailsLogEntry.isGammaTimeout()) {
                numberOfGammaTimeouts++;
                return -1;
            }

            // * if Gamma acyclic:
            Gamma gamma = gammaSat.getGamma();
            gamma.createGammaGraph();

            if (!gamma.isCyclic()) {

                Set<GammaTuple> gammaTupleSetWithoutSpurious = new HashSet<>();
                outer:
                for (GammaTuple tuple : gamma.getGammaTuples()) {
                    if (tuple.getConcept().toString().contains("A_exRC")) {
                        continue;
                    }
                    for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                        if (concept.toString().contains("A_exRC")) {
                            continue outer;
                        }
                    }
                    gammaTupleSetWithoutSpurious.add(tuple);
                }
                Gamma gammaWithoutSpurious = new Gamma(gammaTupleSetWithoutSpurious, startConcept);

                if (verbose) {
                    System.out.println("Concept " + startConcept
                            + " is FO-rewritable (Gamma acyclic, size " + gammaTupleSetWithoutSpurious.size() + ").");
                }
                GammaToDatalogRewriter dr = new GammaToDatalogRewriter(gammaWithoutSpurious, startConcept, context);
                this.resultingProgram = dr.gammaToDatalogProgram();
//                    System.out.println(dr.gammaToDatalogStrings());

                return 0;
            }

            // Increase number of Gamma cycles only once:
            if (!detailsLogEntry.isCyclicGamma()) {
                detailsLogEntry.setCyclicGamma(true);
                numberOfGammaCycles++;
            }

            detailsLogEntry.setNoOfGammaTuplesOnCycle(gamma.numberOfTuplesOnCycle());

            if (verbose) {
                System.out.println("Gamma graph:");
                System.out.println("  Number of vertices:     " + gamma.getGammaGraph().vertexSet().size());
                System.out.println("  Number of edges:        " + gamma.getGammaGraph().edgeSet().size());
                //System.out.print("Distr. of ex. restr. in S: ");
                //System.out.println("    " + countExistentials(gamma.getGammaTuples()));
                System.out.println("  # of vert. lying on cycle: " + gamma.numberOfTuplesOnCycle());
                System.out.println("  distr. of their ex. restr.:");
                System.out.println("    " + countExistentials(gamma.getTuplesOnCycle()));
                //countSimpleCycles(graph);  // VERY costly!!
                System.out.println();
            }
            if (DEBUG) {
                System.out.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                System.out.println("GAMMA:\n");
                System.out.println(GammaSaturator.shortenIDs(gamma.toString()));
                System.out.println("\n-------\n\n");
            }

            // * Saturate Omega:
            OmegaSaturator omegaSat = new OmegaSaturator(gamma, detailsLogEntry, startConcept, context);
            usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
            try {
                omegaSat.saturate(CQContext.getINSTANCE().getTimeout() - usedUntilNow);
            } catch (OutOfMemoryError e) {
                System.err.println(e);
                omegaSat.dumpInternals();
                System.exit(-1);
            }
            Omega omega = omegaSat.getOmega();

            // * if Omega timeout:
            if (detailsLogEntry.isOmegaTimeout()) {
                numberOfOmegaTimeouts++;
                return -2;
            }

            detailsLogEntry.setNoOfOmegaTuples(omega.getOmegaTuples().size());

            // * if Omega cyclic:
            DirectedGraph<OmegaTuple, DefaultEdge> omegaGraph = omegaSat.getOmegaGraph();
            OmegaLoopChecker omegaLoopChecker = new OmegaLoopChecker(startConcept, omegaGraph, omegaSat);
            // To construct the Datalog rewriting, we only need the component 
            // of the graph that is reachable from root tuples. Note that 
            // we actually do the check for Omega loops here and get the 
            // reachable component for free.
            boolean omegaLoopFree = omegaLoopChecker.retainReachableComponentOrFail();
            if (!omegaLoopFree) {
                if (DEBUG) {
                    System.out.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                    System.out.println("OMEGA\n");
                    System.out.println(GammaSaturator.shortenIDs(omega.toString()));
                    System.out.println("\n-------\n\n");
                }
                if (CQContext.getINSTANCE().isVerbose()) {
                    System.out.println("Concept " + startConcept
                            + " is not FO-rewritable (Omega cyclic).");
                }
                numberOfGammaCycleFOno++;
                return 1;
            }

            // * if Omega acyclic and Gamma not premature:
            if (omegaLoopFree && !gammaSat.isPrematureReturnBecauseOfCycle()) {
                if (DEBUG) {
                    System.out.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                    System.out.println("OMEGA\n");
                    System.out.println(GammaSaturator.shortenIDs(omega.toString()));
                    System.out.println("\n-------\n\n");
                }
                if (verbose) {
                    System.out.println("Concept " + startConcept + " is FO-rewritable (Omega cancelling).\n");
                }
                OmegaToDatalogRewriter dr = new OmegaToDatalogRewriter(omegaSat, startConcept, context);
                //System.out.println(dr.omegaToDatalog());
                this.resultingProgram = dr.omegaToDatalogProgram();

                numberOfGammaCycleFOyes++;
                return 0;
            }

            // * Resume Gamma saturation:
            usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
            gammaSat.resumeSaturation(CQContext.getINSTANCE().getTimeout() - usedUntilNow);

        } // while true
    }

    /**
     * This gets all Gamma tuples that lie on a cycle.
     *
     * @param setOfTuplesOnCycle
     * @return
     */
    public static Map<Integer, Integer> countExistentials(Set<GammaTuple> setOfTuplesOnCycle) {

        Map<Integer, Integer> stat = new HashMap<>();

        for (GammaTuple tuple : setOfTuplesOnCycle) {

//            System.err.print("[DEBUG GammaAndOmegaCaller] ");
//            System.err.println("Gamma tuple on cycle: " + tuple);
            int numberOfExistentials = 0;
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                if (concept instanceof OWLObjectSomeValuesFrom) {
                    numberOfExistentials++;
                }
            }
            if (!stat.containsKey(numberOfExistentials)) {
                stat.put(numberOfExistentials, 1);
            } else {
                stat.put(numberOfExistentials, stat.get(numberOfExistentials) + 1);
            }
        }
        return stat;
    }

}
