/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.datalog;

import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogProgram;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogRule;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogClassAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogRoleAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogVariable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;

/**
 *
 * @author hansen
 */
public class GammaToDatalogRewriter {

    Gamma gamma;
    OWLClass startConcept;
    Context context;

    /**
     * Provides methods to produce a Datalog rewriting from an acyclic set of
     * Gamma tuples.
     *
     * @param gamma
     * @param context
     * @param startConcept
     */
    public GammaToDatalogRewriter(Gamma gamma, OWLClass startConcept, Context context) {
        this.gamma = gamma;
        this.startConcept = startConcept;
        this.context = context;
    }

    /**
     * Perform rewriting from set of Gamma tuples into Datalog rules. In the
     * paper, we get a disjunctive Datalog rule from a Gamma tuple (C, S):
     *
     * P_C(x) :- \bigwedge_{A \in S} A(x) \wedge \bigwedge_{\exists r.D \in S}
     * \bigvee_{\Tmc\models s\sqsubseteq r, s\in \Sigma} ( s(x,y_{r,D}) \wedge
     * P_D(y_{r,D}) )
     *
     * Now we introduce auxiliary IDB predicates P_{C, \exists r.D} and get the
     * following:
     *
     * P_C(x) :- \bigwedge_{A \in S} A(x) \wedge \bigwedge_{\exists r.D \in S}
     * P_{C, \exists r.D}(x)
     *
     * P_{C, \exists r.D}(x) :- s(x, y_{r,D}) \wedge P_D(y_{r,D}) for all \Tmc
     * \models s \sqsubseteq r
     *
     * @return
     */
    public DatalogProgram gammaToDatalogProgram() {

        DatalogVariable x0 = new DatalogVariable(0);
        DatalogVariable x1 = new DatalogVariable(1);

        Set<DatalogRule> setOfRules = new HashSet<>();

        // Start by analysing the C parts of the Gamma tuples.
        // Build up a mapping   GoalConcept --> "Q",
        //                     OtherConcept --> "AUX$1",
        //                    OtherConcept2 --> "AUX$2".
        Map<OWLClassExpression, OWLClass> idbMap = new HashMap<>();
        int idbCounter = 1;

        for (GammaTuple g : gamma.getGammaTuples()) {
            OWLClassExpression con = g.getConcept();
            if (!idbMap.containsKey(con)) {
                if (con == startConcept) {
//                    idbMap.put(con, "Q");
                    idbMap.put(con, startConcept);
                } else {
                    // Create Gamma filler name on the fly:
                    String gammaFiString = "GammaFi$" + idbCounter++;
                    OWLClass gammaFiCls = OWLManager.createOWLOntologyManager()
                            .getOWLDataFactory().getOWLClass(gammaFiString, CQContext.getINSTANCE().getPm());
                    idbMap.put(con, gammaFiCls);
                }
            }
        }

        // We will need another set of IDBs to represent existentials in
        // S sets.
        Map<OWLClassExpression, String> sExIdbMap = new HashMap<>();

        // Now process all Gamma tuples.
        for (GammaTuple g : gamma.getGammaTuples()) {

            // Let (C, S) be our Gamma tuple.
            // Now construct the C IDB part (lhs).
            DatalogAtom headAtom = new DatalogClassAtom(idbMap.get(g.getConcept()), x0);
//            result += idbMap.get(g.getConcept()) + "(?0)  <-  ";
            Set<DatalogAtom> bodyAtoms = new HashSet<>();

            for (OWLClassExpression con : g.getSetOfConcepts()) {
                if (con instanceof OWLClass) {
                    OWLClass conCls = (OWLClass) con;
                    DatalogClassAtom newClsAtom = new DatalogClassAtom(conCls, x0);
                    bodyAtoms.add(newClsAtom);
//                    result += StringTools.getFragment(con.asOWLClass().getIRI()) + "(?0), ";

                } else if (con instanceof OWLObjectSomeValuesFrom) {
                    // Let's see if we already have an IDB for this:
                    if (idbMap.containsKey(con)) {
                        DatalogClassAtom newClsAtom = new DatalogClassAtom(idbMap.get(con), x0);
                        bodyAtoms.add(newClsAtom);
//                        result += idbMap.get(con) + "(?0), ";

                    } else {
                        // Generate unique S existential IDB predicate on-the-fly:
                        String sExIdbString = "GammaEx$" + idbCounter++;
                        OWLClass sExIdbCls = OWLManager.createOWLOntologyManager()
                                .getOWLDataFactory().getOWLClass(sExIdbString, CQContext.getINSTANCE().getPm());
                        idbMap.put(con, sExIdbCls);

                        DatalogClassAtom newClsAtom = new DatalogClassAtom(sExIdbCls, x0);
                        bodyAtoms.add(newClsAtom);
//                        result += sExIdbName + "(?0), ";
                        // Add additional IDB rules for this IDB:
                        OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) con;
                        for (OWLObjectPropertyExpression subrole
                                : context.getReasoning().getEntailingRoles(ex.getProperty())) {

                            DatalogAtom addHeadAtom = new DatalogClassAtom(sExIdbCls, x0);
                            DatalogAtom addBodyAtom1 = new DatalogRoleAtom(subrole.asOWLObjectProperty(), x0, x1);
                            DatalogAtom addBodyAtom2 = new DatalogClassAtom(idbMap.get(((OWLObjectSomeValuesFrom) con).getFiller()), x1);
                            Set<DatalogAtom> addBodyAtoms = Sets.newHashSet(addBodyAtom1, addBodyAtom2);
                            DatalogRule additionalRule = new DatalogRule(addHeadAtom, addBodyAtoms);
                            setOfRules.add(additionalRule);
//                            additionalIDBRules += sExIdbName + "(?0)  <-  "
//                                    + StringTools.getFragment(subrole.asOWLObjectProperty().getIRI()) + "(?0, ?1), "
//                                    + idbMap.get(((OWLObjectSomeValuesFrom) con).getFiller()) + "(?1)\n";
                        }
                    }
                }
            }
            // Check if we have A(x) <- A(x) rule:
            if (bodyAtoms.size() == 1) {
                DatalogAtom singleBodyAtom = bodyAtoms.iterator().next();
                if (singleBodyAtom instanceof DatalogClassAtom) {
                    DatalogClassAtom singleClassAtom = (DatalogClassAtom) singleBodyAtom;
                    if (singleClassAtom.equals(headAtom)) {
                        continue;
                    }
                }
            }
            DatalogRule newRule = new DatalogRule(headAtom, bodyAtoms);
            setOfRules.add(newRule);
        }
        DatalogProgram resultingProgram = new DatalogProgram(setOfRules);
        return resultingProgram;
    }

}
