/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.datalog;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.context.OmegaContext;
import de.unibremen.informatik.tdki.grind.omegaconstruction.Omega;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaSaturator;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaSaturator.OmegaTupleRolePair;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaTuple;
import de.unibremen.informatik.tdki.grind.utilities.StringTools;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogProgram;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogRule;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogClassAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogRoleAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogVariable;
import java.util.ArrayDeque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Queue;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.traverse.DepthFirstIterator;
import org.jgrapht.traverse.GraphIterator;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;

/**
 *
 * @author hansen
 */
public class OmegaToDatalogRewriter {

    private final Context context;
    private final OmegaContext omegaContext;
    private final OWLClassExpression startConcept;
    private final OmegaSaturator omegaSat;
    private final Omega omega;
    private final Map<OmegaTuple, Triple> omegaTupleToTriple = new HashMap<>();
    private final Multimap<Triple, OmegaTupleAndOneWitnessList> tripleToOmegaTupleAndOneWitnessList;
    private final Queue<OmegaTupleAndOneWitnessList> setOfOmegaTuplesAndOneWitnessList = new ArrayDeque<>();
    private final Set<OmegaTuple> existentialOmegaTuples = new HashSet<>();
    private final DirectedGraph<Triple, DefaultEdge> tripleGraph
            = new DefaultDirectedGraph<>(DefaultEdge.class);
    private final Set<Triple> rootTriples = new HashSet<>();

    private final Map<Triple, String> tripleToName = new HashMap<>();
    int idbCounter = 0;

    /**
     * Provides methods to produce a Datalog rewriting from an acyclic set of
     * Omega tuples.
     *
     * @param omegaSat
     * @param context
     * @param startConcept
     */
    public OmegaToDatalogRewriter(OmegaSaturator omegaSat, OWLClassExpression startConcept, Context context) {
        this.omegaSat = omegaSat;
        this.omega = omegaSat.getOmega();
        this.startConcept = startConcept;
        this.context = context;
        this.omegaContext = context.getOmegaContext();
        tripleToOmegaTupleAndOneWitnessList = HashMultimap.create();
    }

    /**
     * Contruct a non-recursive monadic Datalog rewriting from an acyclic Omega
     * set. We proceed by building up a data structure of IBD triple, i.e.
     * objects of the form (C, con, XCON).
     *
     * @return
     */
    public DatalogProgram omegaToDatalogProgram() {

        // Separate into OmegaTuples without existentials (do right now) and
        // ones with ex. (put into queue):
        for (OmegaTuple omegaTuple : omega.getOmegaTuples()) {
            boolean onlyCNs = true;
            for (OWLClassExpression concept : omegaTuple.getSetOfConcepts()) {
                if (concept instanceof OWLObjectSomeValuesFrom) {
                    onlyCNs = false;
                    break;
                }
            }
            if (onlyCNs) {
                // If an OmegaTuple has no ex. in S, construct a Triple
                // from it, using null as XCON, and update Maps:
                Triple triple = new Triple(omegaTuple, null);
                tripleGraph.addVertex(triple);
                tripleToOmegaTupleAndOneWitnessList.put(triple, new OmegaTupleAndOneWitnessList(omegaTuple, null));
                omegaTupleToTriple.put(omegaTuple, triple);
            } else {
                // Otherwise we throw it onto the queue to process later:
                existentialOmegaTuples.add(omegaTuple);
            }
        }

        // We build up a set that contains all Objects OmegaTuple, oneWitnessList.
        buildQueueOfOmegaTuplesAndOneWitnessList();
        // All of these have to be processed:

        outerLoop:
        while (!setOfOmegaTuplesAndOneWitnessList.isEmpty()) {
            OmegaTupleAndOneWitnessList currentOmegaTupleAndOneWitnessList
                    = setOfOmegaTuplesAndOneWitnessList.remove();

            // Check that we already have a predicate for this OmegaTuple,
            // otherwise throw the Object back onto the queue:
            OmegaSaturator.WitnessListPlusChosenSuccessor listOfWitnesses = currentOmegaTupleAndOneWitnessList.getListOfWitnesses();

            Set<Triple> witnessTriplesForNewTriple = new HashSet<>();
            for (OmegaTupleRolePair witness : listOfWitnesses.getWitnessList()) {
                if (!omegaTupleToTriple.containsKey(witness.getOmegaTuple())) {
                    setOfOmegaTuplesAndOneWitnessList.add(currentOmegaTupleAndOneWitnessList);
                    continue outerLoop;
                }
                witnessTriplesForNewTriple.add(omegaTupleToTriple.get(witness.getOmegaTuple()));
            }

            // Now we know we have Triples for all witnesses.
            // Go on and build up a new Triple:
            OmegaTuple currentOmegaTuple = currentOmegaTupleAndOneWitnessList.getOmegaTuple();
            Set<Set<OWLClassExpression>> XCON = new HashSet<>();

            // In every M, we have (S_t \cap N_C):
            Set<OWLClassExpression> conceptnames = getConceptnames(currentOmegaTuple.getSetOfConcepts());

            // Built up XCON:
            for (int j = 0; j < listOfWitnesses.getWitnessList().size(); j++) {

                OmegaTupleRolePair witness = listOfWitnesses.getWitnessList().get(j);
                Set<Set<OWLClassExpression>> witnessXCON = omegaTupleToTriple.get(witness.getOmegaTuple()).getXCON();

                // XCON of this witness might be null ( "{--}" ) or empty.
                // Only take \ex r.(con) of the other witnesses:
                if ((witnessXCON == null) || (witnessXCON.isEmpty())) {
                    Set<OWLClassExpression> M = new HashSet<>();
                    M.addAll(conceptnames);
                    for (int i = 0; i < listOfWitnesses.getWitnessList().size(); i++) {
                        if (i == j) {
                            continue;
                        }
                        OmegaTupleRolePair otherWitness = listOfWitnesses.getWitnessList().get(i);
                        addExistsRConT(otherWitness.getOmegaTuple().getCon(), otherWitness.getRole(), M);
                    }
                    XCON.add(omegaContext.closeUnderConsequence(M));
                } // XCON contains some elements:
                else {
                    for (Set<OWLClassExpression> xcon : witnessXCON) {

                        Set<OWLClassExpression> M = new HashSet<>();
                        M.addAll(conceptnames);

                        // Add \ex r.(xcon) of this xcon element:
                        addExistsRConT(xcon, witness.getRole(), M);

                        // Add \ex r.(con) for other witnesses:
                        for (int i = 0; i < listOfWitnesses.getWitnessList().size(); i++) {
                            if (i == j) {
                                continue;
                            }
                            OmegaTupleRolePair otherWitness = listOfWitnesses.getWitnessList().get(i);
                            addExistsRConT(otherWitness.getOmegaTuple().getCon(), otherWitness.getRole(), M);
                        }
                        XCON.add(omegaContext.closeUnderConsequence(M));
                    }
                }
            }

            Triple newTriple = new Triple(currentOmegaTuple, XCON);

            // Add new Triple and edges to witnessTriples to Triple Graph:
            tripleGraph.addVertex(newTriple);
            for (Triple witnessTriple : witnessTriplesForNewTriple) {
                tripleGraph.addEdge(newTriple, witnessTriple);
            }

            omegaTupleToTriple.put(currentOmegaTuple, newTriple);
            tripleToOmegaTupleAndOneWitnessList.put(newTriple, currentOmegaTupleAndOneWitnessList);
        }

        DatalogProgram resultProgram = generateDatalogRewritingFromTripleSet();
        return resultProgram;
    }

    private DatalogProgram generateDatalogRewritingFromTripleSet() {

        DatalogVariable x0 = new DatalogVariable(0);
        DatalogVariable x1 = new DatalogVariable(1);

        Set<DatalogRule> setOfRules = new HashSet<>();

        // Restrict Triple Graph to reachable components from root Triples:
        retainReachableTripleGraphComponent();

        // Generate names for IDB predicates:
        mapTriplesToIDBNames();

        // Iterate over triples and produce rules for each of them:
        for (Triple triple : tripleGraph.vertexSet()) {

            for (OmegaTupleAndOneWitnessList omegaTupleAndOneWitnessList
                    : tripleToOmegaTupleAndOneWitnessList.get(triple)) {

                //result += tripleToName.get(triple) + "(?0)  <-  ";
                OWLClass tripleCls = OWLManager.getOWLDataFactory().getOWLClass(tripleToName.get(triple), CQContext.getINSTANCE().getPm());
                DatalogAtom head = new DatalogClassAtom(tripleCls, x0);

                OmegaTuple omegaTuple = omegaTupleAndOneWitnessList.getOmegaTuple();

                // First do all conceptnames in S:
                Set<DatalogAtom> bodyAtoms = new HashSet<>();
                for (OWLClassExpression concept : omegaTuple.getSetOfConcepts()) {
                    if (concept instanceof OWLClass) {
                        //result += concept.asOWLClass() + "(?0), ";
                        DatalogAtom bodyAtom = new DatalogClassAtom(concept.asOWLClass(), x0);
                        bodyAtoms.add(bodyAtom);
                    }
                }

                // Now we have to get the predicates for the witnesses in the
                // witness list:
                if (omegaTupleAndOneWitnessList.getListOfWitnesses() != null) {
                    for (OmegaTupleRolePair omegaTupleRolePair
                            : omegaTupleAndOneWitnessList.getListOfWitnesses().getWitnessList()) {

                        //result += omegaTupleRolePair.getRole().asOWLObjectProperty() + "(?0, ?1), ";
                        DatalogAtom bodyAtom = new DatalogRoleAtom(omegaTupleRolePair.getRole().asOWLObjectProperty(), x0, x1);
                        bodyAtoms.add(bodyAtom);

                        Triple witnessTriple = omegaTupleToTriple.get(omegaTupleRolePair.getOmegaTuple());

                        //result += tripleToName.get(witnessTriple) + "(?1), ";
                        tripleCls = OWLManager.getOWLDataFactory()
                                .getOWLClass(tripleToName.get(witnessTriple), CQContext.getINSTANCE().getPm());
                        bodyAtom = new DatalogClassAtom(tripleCls, x1);
                        bodyAtoms.add(bodyAtom);
                    }
                }

                DatalogRule rule = new DatalogRule(head, bodyAtoms);
                setOfRules.add(rule);
            }
        }

        DatalogProgram resultProgram = new DatalogProgram(setOfRules);
        return resultProgram;
    }

    private void mapTriplesToIDBNames() {

        for (Triple triple : tripleGraph.vertexSet()) {
            String name;
            if (rootTriples.contains(triple)) {
                name = startConcept.toString();
//                name = "Q";
            } else {
                name = "Omega$" + idbCounter++;
            }
            tripleToName.put(triple, name);
        }
    }

    /**
     * A Triple (C,con,XCON) is a root triple if C = A0, A0 in con and A0 not in
     * xcon for all xcon in XCON.
     */
    private void calculateRootTriples() {

        nextTriple:
        for (Triple triple : tripleGraph.vertexSet()) {
            if (triple.getConcept().equals(startConcept)
                    && triple.getCon().contains(startConcept)) {
                if (triple.getXCON() == null
                        || triple.getXCON().isEmpty()) {
                    rootTriples.add(triple);
                } else {
                    for (Set<OWLClassExpression> xcon : triple.getXCON()) {
                        if (xcon.contains(startConcept)) {
                            continue nextTriple;
                        }
                    }
                    rootTriples.add(triple);
                }
            }
        }
    }

    public void retainReachableTripleGraphComponent() {

        Set<Triple> verticesToRetain = new HashSet<>();

        // Check for all the root tuples in Omega:
        calculateRootTriples();
        for (Triple rootTriple : rootTriples) {

            GraphIterator<Triple, DefaultEdge> iterator
                    = new DepthFirstIterator<>(tripleGraph, rootTriple);
            while (iterator.hasNext()) {
                verticesToRetain.add(iterator.next());
            }
        }

        // Now we did a complete dfs on the nodes reachable from root triples.
        // The vertices we visited are stored in verticesToRetain. 
        // Delete the others from graph:
        Set<Triple> verticesToDelete = new HashSet<>(tripleGraph.vertexSet());
        verticesToDelete.removeAll(verticesToRetain);
        tripleGraph.removeAllVertices(verticesToDelete);
    }

    /**
     * Add \exists relevantRole.C for all C in cxon.
     *
     * @param cxon The consequences to add via relevantRole.
     * @param relevantRole
     * @param setM This set will be added to!
     */
    private void addExistsRConT(Set<OWLClassExpression> cxon,
            OWLObjectPropertyExpression relevantRole, Set<OWLClassExpression> setM) {

        // If we have to construct \exists r.--, just add \top if it is in sub(T).
        if (cxon == null) {
            if (context.getOriginalOnt().getSubconcepts().contains(context.topConcept())) {
                setM.add(context.topConcept());
            }
            return;
        }
        for (OWLClassExpression g : cxon) {

            // Put all exists [superrole of r].G into con! This will also help
            // if exists r.G is not in sub(T)!
            for (OWLObjectPropertyExpression superrole : context.getReasoning().getEntailedRoles(relevantRole)) {
                OWLObjectSomeValuesFrom existsRG
                        = OWLManager.getOWLDataFactory().getOWLObjectSomeValuesFrom(superrole, g);
                if (omegaContext.getContext().getOriginalOnt().getSubconcepts().contains(existsRG)) {
                    setM.add(existsRG);
                }
            }
        }
    }

    private static Set<OWLClassExpression> getConceptnames(Set<OWLClassExpression> concepts) {
        Set<OWLClassExpression> conceptnames = new HashSet<>();
        for (OWLClassExpression concept : concepts) {
            if (concept instanceof OWLClass) {
                conceptnames.add(concept);
            }
        }
        return conceptnames;
    }

    private void buildQueueOfOmegaTuplesAndOneWitnessList() {
        for (OmegaTuple tuple : existentialOmegaTuples) {
            Set<OmegaSaturator.WitnessListPlusChosenSuccessor> setOfListOfWitnesses
                    = omegaSat.getOmegaTupleToSetOfListOfWitnessesPlusSucc().get(tuple);
            for (OmegaSaturator.WitnessListPlusChosenSuccessor listOfWitnesses : setOfListOfWitnesses) {
                OmegaTupleAndOneWitnessList omegaTupleAndOneWitnessList
                        = new OmegaTupleAndOneWitnessList(tuple, listOfWitnesses);
                setOfOmegaTuplesAndOneWitnessList.add(omegaTupleAndOneWitnessList);

            }
        }
    }

    private class Triple {

        private OWLClassExpression concept;
        private Set<OWLClassExpression> con;
        private Set<Set<OWLClassExpression>> XCON;

        public Triple(OmegaTuple tuple, Set<Set<OWLClassExpression>> XCON) {
            this.concept = tuple.getConcept();
            this.con = tuple.getCon();
            this.XCON = XCON;
        }

        public OWLClassExpression getConcept() {
            return concept;
        }

        public Set<OWLClassExpression> getCon() {
            return con;
        }

        public Set<Set<OWLClassExpression>> getXCON() {
            return XCON;
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 97 * hash + Objects.hashCode(this.concept);
            hash = 97 * hash + Objects.hashCode(this.con);
            hash = 97 * hash + Objects.hashCode(this.XCON);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final Triple other = (Triple) obj;
            if (!Objects.equals(this.concept, other.concept)) {
                return false;
            }
            if (!Objects.equals(this.con, other.con)) {
                return false;
            }
            if (!Objects.equals(this.XCON, other.XCON)) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "Triple{" + "concept=" + concept + ", con=" + con + ", XCON=" + XCON + "}\n";
        }

    }

    private class OmegaTupleAndOneWitnessList {

        private final OmegaTuple omegaTuple;
        private final OmegaSaturator.WitnessListPlusChosenSuccessor listOfWitnessesPlusSucc;

        public OmegaTupleAndOneWitnessList(OmegaTuple omegaTuple, OmegaSaturator.WitnessListPlusChosenSuccessor listOfWitnesses) {
            this.omegaTuple = omegaTuple;
            this.listOfWitnessesPlusSucc = listOfWitnesses;
        }

        public OmegaTuple getOmegaTuple() {
            return omegaTuple;
        }

        public OmegaSaturator.WitnessListPlusChosenSuccessor getListOfWitnesses() {
            return listOfWitnessesPlusSucc;
        }

        @Override
        public int hashCode() {
            int hash = 7;
            hash = 53 * hash + Objects.hashCode(this.omegaTuple);
            hash = 53 * hash + Objects.hashCode(this.listOfWitnessesPlusSucc);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final OmegaTupleAndOneWitnessList other = (OmegaTupleAndOneWitnessList) obj;
            if (!Objects.equals(this.omegaTuple, other.omegaTuple)) {
                return false;
            }
            if (!Objects.equals(this.listOfWitnessesPlusSucc, other.listOfWitnessesPlusSucc)) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "OmegaTupleAndOneWitnessList{" + "omegaTuple=" + omegaTuple + ", listOfWitnessesPlusSucc=" + listOfWitnessesPlusSucc + '}';
        }

    }

}
