/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.gammaconstruction;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.statistics.DetailsLogEntry;
import de.unibremen.informatik.tdki.grind.utilities.ELUtils;
import de.unibremen.informatik.tdki.grind.utilities.SetTrie;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.alg.DijkstraShortestPath;
import org.jgrapht.traverse.DepthFirstIterator;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;

/**
 *
 * @author hansen
 */
public class GammaSaturator {

    // Use this to speed up things when doing real work:
    private static final boolean DEBUG = false;
    
    // Decide whether we want to do the instantaneous return on finding 
    // (C, {ex r.C}) tuple:
    private static final boolean DO_REFLEXIVITY_PREMATURITY_CHECK = false;

    // After e.g. 500 iterations, do one cyclicity check on Gamma!
    final int PREMATURE_CHECK_THRESHOLD = 1000;

    private final OWLClass startConcept;
    private Map<OWLClassExpression, SetTrie<OWLClassExpression>> tupleQ;
    private Map<OWLClassExpression, SetTrie<OWLClassExpression>> processedTuples;
    private final Context context;
    private long startTime, t;
    private int debugCounter;
    private DetailsLogEntry detailsLogEntry;
    private boolean prematureReturnBecauseOfCycle = false;
    private GammaTuple reflexiveLoopingTuple = null;
    private Gamma gamma;

    public Gamma getGamma() {
        return gamma;
    }

    public GammaSaturator(OWLClass startConcept, DetailsLogEntry detailsLogEntry, Context context) {
        this.startConcept = startConcept;
        this.context = context;
        this.detailsLogEntry = detailsLogEntry;
    }

    public GammaSaturator(OWLClass startConcept, Context context) {
        this(startConcept, new DetailsLogEntry(), context);
    }

    public boolean isPrematureReturnBecauseOfCycle() {
        return prematureReturnBecauseOfCycle;
    }

    private boolean safeAddToMap(Map<OWLClassExpression, SetTrie<OWLClassExpression>> map, OWLClassExpression concept, SortedSet<OWLClassExpression> set) {
        if (map.containsKey(concept)) {
            return map.get(concept).insert(set);
        } else {
            SetTrie<OWLClassExpression> trie = new SetTrie<>(context.getOriginalOnt().getComparator());
            trie.insert(set);
            map.put(concept, trie);
            return true;
        }
    }

    private void safeRemoveSupersets(Map<OWLClassExpression, SetTrie<OWLClassExpression>> map, OWLClassExpression concept, SortedSet<OWLClassExpression> set) {
        if (map.containsKey(concept)) {
            map.get(concept).removeSupersets(set);
        }
    }

    public void saturate() {
        saturate(Long.MAX_VALUE);
    }

    /**
     * Generates a set of GammaTuples from processed. To be called when the
     * queue has been emptied.
     *
     * @return
     */
    private Gamma generateGammaFromProcessed() {
        Gamma resultingGamma = new Gamma(startConcept);

        detailsLogEntry.resetGammaSSetCardinalities();
        detailsLogEntry.setNoOfDifferentGammaCConcepts(processedTuples.keySet().size());

        for (Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entryCAndAllSSets : processedTuples.entrySet()) {
            SetTrie<OWLClassExpression> trieOfAllSSets = entryCAndAllSSets.getValue();
            Set<Set<OWLClassExpression>> setOfSets = trieOfAllSSets.toSetOfSets();
            for (Set<OWLClassExpression> unorderedSetOfConcepts : setOfSets) {
                SortedSet sSet = new TreeSet<>(context.getOriginalOnt().getComparator());
                sSet.addAll(unorderedSetOfConcepts);
                GammaTuple extractedTuple = new GammaTuple(entryCAndAllSSets.getKey(), sSet);
                resultingGamma.addTuple(extractedTuple);
                detailsLogEntry.increaseGammaSSetWithCardinality(extractedTuple.getSetOfConcepts().size());
            }
        }

        detailsLogEntry.setGammaTimeout(false);
        detailsLogEntry.setNoOfGammaTuples(resultingGamma.size());
        return resultingGamma;
    }

    /**
     * Combines tupleQ and processedTuples into one Map and returns the
     * resulting Gamma. Useful when immediateReturn'ing b/c of a tuple (C,ex
     * r.C).
     *
     * @return
     */
    private Gamma generateGammaFromQAndProcessed() {

        Gamma resultingGamma = new Gamma(startConcept);

        detailsLogEntry.resetGammaSSetCardinalities();
        Set<OWLClassExpression> countingSet = new HashSet<>();
        countingSet.addAll(tupleQ.keySet());
        countingSet.addAll(processedTuples.keySet());
        detailsLogEntry.setNoOfDifferentGammaCConcepts(countingSet.size());

        List<Map<OWLClassExpression, SetTrie<OWLClassExpression>>> mapList
                = Arrays.asList(tupleQ, processedTuples);
        for (Map<OWLClassExpression, SetTrie<OWLClassExpression>> qOrProcessed : mapList) {
            for (Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entryCAndAllSSets : qOrProcessed.entrySet()) {
                SetTrie<OWLClassExpression> trieOfAllSSets = entryCAndAllSSets.getValue();
                Set<Set<OWLClassExpression>> setOfSets = trieOfAllSSets.toSetOfSets();
                for (Set<OWLClassExpression> unorderedSetOfConcepts : setOfSets) {
                    SortedSet sSet = new TreeSet<>(context.getOriginalOnt().getComparator());
                    sSet.addAll(unorderedSetOfConcepts);
                    GammaTuple extractedTuple = new GammaTuple(entryCAndAllSSets.getKey(), sSet);
                    resultingGamma.addTuple(extractedTuple);
                    detailsLogEntry.increaseGammaSSetWithCardinality(extractedTuple.getSetOfConcepts().size());
                }
            }
        }

        detailsLogEntry.setGammaTimeout(false);
        detailsLogEntry.setNoOfGammaTuples(resultingGamma.size());
        return resultingGamma;
    }

    /**
     * A tuple (A0, S) is a root tuple.
     */
    private Set<GammaTuple> getRootTuples(Gamma gamma) {
        Set<GammaTuple> result = new HashSet<>();
        for (GammaTuple tuple : gamma.getGammaTuples()) {
            if (tuple.getConcept().equals(startConcept)) {
                result.add(tuple);
            }
        }
        return result;
    }

    /**
     * When a (C,ex r.C...) tuple is found, restrict Gamma to a component
     * containing a root tuple whose successor is (C,ex r.C...).
     */
    private Gamma restrictGammaToLoopingPathComponent(Gamma gamma, GammaTuple loopingTuple) {

        GammaGraphCreator ggc = new GammaGraphCreator(gamma, startConcept);
        DirectedGraph gammaGraph = ggc.getGraph();

        Gamma newGamma = new Gamma(startConcept);
        int noOfTouchingRootTuples = 0;

        for (GammaTuple rootTuple : getRootTuples(gamma)) {
            final List<GammaGraphCreator.RoleEdge> pathListOfEdges
                    = DijkstraShortestPath.findPathBetween(gammaGraph, rootTuple, loopingTuple);
            if (pathListOfEdges != null) {
                noOfTouchingRootTuples++;
                // Now we found a root tuple rootTuple whose successor is loopingTuple!
                // Get this root tuple's connected component:
                DepthFirstIterator<GammaTuple, GammaGraphCreator.RoleEdge> dfi
                        = new DepthFirstIterator<>(gammaGraph, rootTuple);
                while (dfi.hasNext()) {
                    newGamma.addTuple(dfi.next());
                }
                return newGamma;
            }
        }

//        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//        System.err.println("No. of touching root tuples: " + noOfTouchingRootTuples);
        if (noOfTouchingRootTuples > 0) {
            return newGamma;
        }

        // Cannot restrict Gamma to looping path. Something's wrong.
        // Leave gamma intact.
//        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//        System.err.println("Shouldn't happen!");
        return gamma;
    }

    /**
     * This applies the Gamma rules (1) and (2) to saturate the set Gamma.
     *
     * @param timeoutPeriod
     * @return
     */
    public void saturate(long timeoutPeriod) {

        if (DEBUG) {
            startTime = System.currentTimeMillis();
        }
        // Open a Queue with tuples to be processed.
        tupleQ = new HashMap<>();

        processedTuples = new HashMap<>();
        
        // Add (A0, {A0}) as first tuple:
        SortedSet set = new TreeSet<>(context.getOriginalOnt().getComparator());
        set.add(startConcept);
        GammaTuple initialTuple = new GammaTuple(startConcept, set);
        addToQIfNotProcessed(initialTuple);

        if (DEBUG) {
            t = System.currentTimeMillis();
            debugCounter = 0;
        }
        long timeoutBegin = System.currentTimeMillis();

        // This will count the number of iterations of the outer loop:
        int iterationsOfWhileLoop = 0;

        // do while Q is not empty
        while (!tupleQ.isEmpty()) {

            iterationsOfWhileLoop++;

            // The following generates debugging output:
//            if (debug) {
//                if ((System.currentTimeMillis() - t) > 500) {
//                    outputShortDebuggingInfo();
//                    t = System.currentTimeMillis();
//                    debugCounter++;
//                    if (debugCounter > 2) {
//                        //outputDebuggingInfo();
//                        debugCounter = 0;
//                    }
//                }
//            }

            // Now we have different heuristics that determine which tuple
            // out of tupleQ to process next:
            GammaTuple tuple = getMinFromQ();
            
            // Take the tuple from Q and put it in processedTuples.
            safeAddToMap(processedTuples, tuple.getConcept(), tuple.getSetOfConcepts());

            applyRule(tuple);

            // Check if applyRule found a (C, {ex r.C}) tuple:
            if (DO_REFLEXIVITY_PREMATURITY_CHECK && prematureReturnBecauseOfCycle) {
                gamma = generateGammaFromQAndProcessed();
                gamma = restrictGammaToLoopingPathComponent(gamma, reflexiveLoopingTuple);
                return;
            }

            long usedUntilNow = System.currentTimeMillis() - timeoutBegin;
            boolean isTimeout = (usedUntilNow >= timeoutPeriod);
            if (isTimeout) {
                detailsLogEntry.setGammaTimeout(true);
                System.out.println("Timed out, Gamma phase!\n");
                return;
            }

            // Periodic Gamma cycle check.
            // Check if we need to do the periodic cyclicity check:
            if (iterationsOfWhileLoop > PREMATURE_CHECK_THRESHOLD) {
                iterationsOfWhileLoop = 0;

//                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                System.err.println("Periodic cyclicity check running...");

                gamma = this.generateGammaFromQAndProcessed();
                GammaGraphCreator ggc = new GammaGraphCreator(gamma, startConcept);
                DirectedGraph gammaGraph = ggc.getGraph();

//                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                System.err.println("Temporary Gamma graph vertices: " + gammaGraph.vertexSet().size());
//                System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                System.err.println("Temporary Gamma graph edges:    " + gammaGraph.edgeSet().size());

                CycleDetector cd = new CycleDetector(gammaGraph);
                boolean gammaCyclic = cd.detectCycles();
                if (gammaCyclic) {
                    prematureReturnBecauseOfCycle = true;
//                    System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                    System.err.println("Found premature Gamma cycle.");

                    Set<GammaTuple> tuplesOnSomeCycle = cd.findCycles();
                    // Take one of these tuples randomly:
                    gamma = restrictGammaToLoopingPathComponent(gamma, tuplesOnSomeCycle.iterator().next());
                    return;
                }
            }
        }
        gamma = generateGammaFromProcessed();
    }

    /**
     * Resume calculation of Gamma, using already existing tupleQ and
     * processedTuples. This happens when we have detected a Gamma, then tried
     * Omega, but Omega didn't find a cycle.
     * @param timeoutPeriod
     */
    public void resumeSaturation(long timeoutPeriod) {

        // We don't do any further premature returns:
        prematureReturnBecauseOfCycle = false;
        
        final boolean doPrematureChecks = false;

        if (DEBUG) {
            startTime = System.currentTimeMillis();
            t = System.currentTimeMillis();
            debugCounter = 0;
        }
        long timeoutBegin = System.currentTimeMillis();

        // This will count the number of iterations of the outer loop:
        int iterationsOfWhileLoop = 0;

        // do while Q is not empty
        while (!tupleQ.isEmpty()) {

            iterationsOfWhileLoop++;

            // The following generates debugging output:
//            if (debug) {
//                if ((System.currentTimeMillis() - t) > 500) {
//                    outputShortDebuggingInfo();
//                    t = System.currentTimeMillis();
//                    debugCounter++;
//                    if (debugCounter > 2) {
//                        //outputDebuggingInfo();
//                        debugCounter = 0;
//                    }
//                }
//            }
            // Now we have different heuristics that determine which tuple
            // out of tupleQ to process next:
            GammaTuple tuple = getMinFromQ();

            // Take the tuple from Q and put it in processedTuples.
            safeAddToMap(processedTuples, tuple.getConcept(), tuple.getSetOfConcepts());

            applyRule(tuple);

            if (doPrematureChecks) {
                // Check if applyRule found a (C,ex r.C) tuple:
                if (prematureReturnBecauseOfCycle) {

                    gamma = generateGammaFromQAndProcessed();
                    gamma = restrictGammaToLoopingPathComponent(gamma, reflexiveLoopingTuple);
                    return;
                }
            }

            long usedUntilNow = System.currentTimeMillis() - timeoutBegin;
            boolean isTimeout = (usedUntilNow >= timeoutPeriod);
            if (isTimeout) {
                detailsLogEntry.setGammaTimeout(true);
                System.out.println("Timed out, Gamma phase!\n");
                return;
            }

            if (doPrematureChecks) {
                // Periodic Gamma cycle check.
                // Check if we need to do the periodic cyclicity check:
                if (iterationsOfWhileLoop > PREMATURE_CHECK_THRESHOLD) {
                    iterationsOfWhileLoop = 0;

//                    System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                    System.err.println("Periodic cyclicity check running...");

                    gamma = this.generateGammaFromQAndProcessed();
                    GammaGraphCreator ggc = new GammaGraphCreator(gamma, startConcept);
                    DirectedGraph gammaGraph = ggc.getGraph();

//                    System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                    System.err.println("Temporary Gamma graph vertices: " + gammaGraph.vertexSet().size());
//                    System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                    System.err.println("Temporary Gamma graph edges:    " + gammaGraph.edgeSet().size());

                    CycleDetector cd = new CycleDetector(gammaGraph);
                    boolean gammaCyclic = cd.detectCycles();
                    if (gammaCyclic) {
                        prematureReturnBecauseOfCycle = true;
//                        System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                        System.err.println("Found premature Gamma cycle.");

                        Set<GammaTuple> tuplesOnSomeCycle = cd.findCycles();
                        // Take one of these tuples randomly:
                        gamma = restrictGammaToLoopingPathComponent(gamma, tuplesOnSomeCycle.iterator().next());
                        return;
                    }

                }
            }

        } // while Q not empty
        gamma = generateGammaFromProcessed();
    }

    /**
     * Heuristics (1): Get first from Q.
     */
    private GammaTuple getFirstFromQ() {

        // Pick first key from tupleQ
        Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entry = tupleQ.entrySet().iterator().next();
        OWLClassExpression C = entry.getKey();
        SetTrie<OWLClassExpression> trie = entry.getValue();
        SortedSet<OWLClassExpression> S = trie.popFirst();
        if (trie.isEmpty()) {
            tupleQ.remove(C);
        }
        return new GammaTuple(C, S);
    }

    private GammaTuple getMinFromQ() {

        // Pick min. key from tupleQ
        Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entry = tupleQ.entrySet().iterator().next();
        OWLClassExpression C = entry.getKey();
        SetTrie<OWLClassExpression> trie = entry.getValue();
        SortedSet<OWLClassExpression> S = trie.popMin();
        if (trie.isEmpty()) {
            tupleQ.remove(C);
        }
        return new GammaTuple(C, S);
    }

    // TODO: Get GLOBAL minimal S.
//    /**
//    * Heuristics (2): Get an entry with a minimal |S|.
//     */
//    private Map.Entry<OWLClassExpression, Set<OWLClassExpression>> getMinimalS() {
//
//        Map.Entry<OWLClassExpression, Set<OWLClassExpression>> nextEntry = null;
//        boolean foundEntry = false;
//        int maxSize = 0;
//        
//        while (!foundEntry) {
//            Iterator<Map.Entry<OWLClassExpression, Set<OWLClassExpression>>> entryIt = 
//                    tupleQ.entries().iterator();
//            do nextEntry = entryIt.next(); 
//            while (nextEntry.getValue().size() > maxSize && entryIt.hasNext());
//            if (nextEntry.getValue().size() == maxSize) foundEntry = true;
//            maxSize++;
//        }
//        return nextEntry;
//    }
    private boolean addToQIfNotProcessed(GammaTuple tuple) {

        // Call shrink optimisation
//        shrink(tuple);
        // TOP concept optimisation -- we remove TOP from S
        // first condition ensures that TOP has a valid integer id in (concept - id) mapping
        if (context.getOriginalOnt().getSubconcepts().contains(context.topConcept())
                && tuple.getSetOfConcepts().contains(context.topConcept())) {
            tuple.getSetOfConcepts().remove(context.topConcept());
        }

        if (processedTuples.containsKey(tuple.getConcept())
                && processedTuples.get(tuple.getConcept()).containsSubsetOf(tuple.getSetOfConcepts())) {
            return false;
        }

        boolean inserted = safeAddToMap(tupleQ, tuple.getConcept(), tuple.getSetOfConcepts());
        if (inserted) {
            safeRemoveSupersets(processedTuples, tuple.getConcept(), tuple.getSetOfConcepts());

            //
            // Heuristics 3: Check whether we have (C, {...\exists r.C,...})
            if (DO_REFLEXIVITY_PREMATURITY_CHECK) {
                OWLClassExpression cConcept = tuple.getConcept();
                for (OWLClassExpression sConcept : tuple.getSetOfConcepts()) {
                    if (sConcept instanceof OWLObjectSomeValuesFrom) {
                        OWLObjectSomeValuesFrom ex = (OWLObjectSomeValuesFrom) sConcept;
                        // Now check whether (C, {ex r.C}):
                        if (ex.getFiller().equals(cConcept)) {
//                            System.err.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
//                            System.err.println("Found (C, {...ex r.C...})!");
                            prematureReturnBecauseOfCycle = true;
                            reflexiveLoopingTuple = tuple;
                            return true;
                        }
                    }
                }
            }

        }
        return inserted;
    }

    private void applyRule(GammaTuple tuple) {
        Set<OWLSubClassOfAxiom> allRelevantAxioms = new HashSet<>();

        // Put all 'relevant' GCIs into set, i.e. axioms with matching rhs.
        // For role hierarchies, this is a bit more complex: If we have a GCI 
        //   D => \exists r.F and C
        // and in the tuple's S set we have \exists s.G, and we know that
        // T \models F => G (from ELK), r => s, then the GCI is relevant
        // and can be applied backwards, so \exists s.G has to be replaced
        // by the tlcs of D. 
        // See Reasoning.java, where the mapping is built.
        for (OWLClassExpression c : tuple.getSetOfConcepts()) {
            if (context.getReasoning().hasRelevantAxiom(c)) {
                allRelevantAxioms.addAll(context.getReasoning().getRelevantAxioms(c));
            }
        }

        // For every matching GCI, we produce Gamma tuples.
        for (OWLSubClassOfAxiom gci : allRelevantAxioms) {
            SortedSet<OWLClassExpression> newSetOfConcepts = new TreeSet<>(context.getOriginalOnt().getComparator());
            newSetOfConcepts.addAll(tuple.getSetOfConcepts());

            for (OWLClassExpression c : ELUtils.getTlcs(gci.getSuperClass())) {
                // First, delete matching concept names.
                if (c instanceof OWLClass) {
                    newSetOfConcepts.remove(c);

                } // Then look for existentials \exists r.F in rhs.
                // Delete all \exists s.G in S with T \models r => s, F => G.
                else {
                    OWLObjectSomeValuesFrom existential = (OWLObjectSomeValuesFrom) c;
                    // Get all entailed roles s and all entailed concepts G:
                    for (OWLObjectPropertyExpression s : context.getReasoning().getEntailedRoles(existential.getProperty())) {
                        for (OWLClassExpression g : context.getReasoning().getEntailedConcepts(existential.getFiller())) {
                            // Construct the concept \exists s.G and see if we can remove it from S.
                            OWLObjectSomeValuesFrom existsSG = context.getFactory().getOWLObjectSomeValuesFrom(s, g);
                            // To avoid NullPointerException when removing, check first.
                            if (context.getOriginalOnt().getSubconcepts().contains(existsSG)) {
                                newSetOfConcepts.remove(existsSG);
                            }
                        }
                    }
                }
            }
            newSetOfConcepts.addAll(ELUtils.getTlcs(gci.getSubClass()));
            GammaTuple newTuple = new GammaTuple(tuple.getConcept(), newSetOfConcepts);
            boolean addToQRetval = addToQIfNotProcessed(newTuple);
            if (prematureReturnBecauseOfCycle) {
                return;
            }
            
            // If we added a tuple, check for its existentials (ex r.G) in S, 
            // and add tuples (G, {ex r.G}) for all of them!
            if (addToQRetval) {
                for (OWLClassExpression c : ELUtils.getSubconcepts(gci.getSubClass())) {
                    if (c instanceof OWLObjectSomeValuesFrom) {
                        OWLObjectSomeValuesFrom existsRG = (OWLObjectSomeValuesFrom) c;
                        newSetOfConcepts = new TreeSet<>(context.getOriginalOnt().getComparator());
                        newSetOfConcepts.addAll(ELUtils.getTlcs(existsRG.getFiller()));
                        newTuple = new GammaTuple(existsRG.getFiller(), newSetOfConcepts);
                        addToQIfNotProcessed(newTuple);
                        if (prematureReturnBecauseOfCycle) {
                            return;
                        }
                    }
                }
            }
        }
    }

    private void outputShortDebuggingInfo() {
        System.out.print("\n    sat()     # of tuples in Q:                   ");
        int numberOfTuples = 0;
        for (Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entry : tupleQ.entrySet()) {
            numberOfTuples += entry.getValue().toSetOfSets().size();
        }
        System.out.println(numberOfTuples);

        //        if (tupleQ.size() == 1) {
//            System.out.println(tupleQ.get(tupleQ.keySet().iterator().next()).toSetOfSets());
//        }
        System.out.print("              # of tuples in processedTuples:     ");
        numberOfTuples = 0;
        for (Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entry : processedTuples.entrySet()) {
            numberOfTuples += entry.getValue().toSetOfSets().size();
        }
        System.out.println(numberOfTuples);

        System.out.print("              time spent saturating this concept: ");
        System.out.printf("%.1f", ((double) (System.currentTimeMillis() - startTime)) / 60000.0);
        System.out.println(" min.");
    }

    private void outputDebuggingInfo() {
        System.out.println("           Q:");
        outputStatistics(tupleQ);
        System.out.println("        proT:");
        outputStatistics(processedTuples);
    }

    private void outputStatistics(Map<OWLClassExpression, SetTrie<OWLClassExpression>> map) {

        int maxSize = 0;
        int maxDepth = 0;
        Multiset<Integer> collectionOfSizes;
        collectionOfSizes = HashMultiset.create();

        Set<OWLClassExpression> unusedSubconcepts = new HashSet<>();

        // Start by declaring all sub(T) as unused.
        unusedSubconcepts.addAll(context.getOriginalOnt().getSubconcepts());

        for (Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entry : map.entrySet()) {
            for (Set<OWLClassExpression> setOfConcepts : entry.getValue().toSetOfSets()) {
                int size = setOfConcepts.size();
                collectionOfSizes.add(size);
                if (size > maxSize) {
                    maxSize = size;
                }

                // Remove used subconcept.
                unusedSubconcepts.removeAll(setOfConcepts);
//                assert(context.getSubconcepts().containsAll(setOfConcepts));

                for (OWLClassExpression concept : setOfConcepts) {
                    int depth = ELUtils.getDepth(concept);
                    if (depth > maxDepth) {
                        maxDepth = depth;
                    }
                }
            }
        }
        System.out.println("              size of biggest S set:              " + maxSize);
        System.out.println("              depth of deepest concept:           " + maxDepth);
        System.out.println("              size of sub(T):                     " + context.getOriginalOnt().getSubconcepts().size());
        System.out.println("              # of concepts used:                 "
                + (context.getOriginalOnt().getSubconcepts().size() - unusedSubconcepts.size()));
        System.out.println("            Distribution of |S| of the tuples:");
        for (int i = 0; i <= maxSize; i++) {
            System.out.print("              size ");
            System.out.printf("%2d", i);
            System.out.printf(", %5d", collectionOfSizes.count(i));
            System.out.println(" occurences.");
        }

        System.out.println("\n            Full dump:");
        for (Map.Entry<OWLClassExpression, SetTrie<OWLClassExpression>> entry : map.entrySet()) {
            for (Set<OWLClassExpression> setOfConcepts : entry.getValue().toSetOfSets()) {
                System.out.print("              " + shortenIDs(entry.getKey().toString()));
                System.out.println(" -- " + shortenIDs(setOfConcepts.toString()));
            }
        }
        System.out.println();
    }

    public static String shortenIDs(String s) {
//        return s.replace("ObjectSomeValuesFrom", "Ex").replace("http://purl.obolibrary.org/obo/FBbi_00000", "");
//        return s.replace("ObjectSomeValuesFrom", "Ex").replace("http://www.opengalen-cohse.org#", "");
        return s.replace("ObjectSomeValuesFrom", "Ex").replace("http://www.informatik.uni-bremen.de/tdki/hansen/", "");
    }
}
