/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.omegaconstruction;

import java.util.HashSet;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.model.OWLClass;

/**
 *
 * @author hansen
 */
public class OmegaLoopChecker {

    private final DirectedGraph<OmegaTuple, DefaultEdge> graph;
    private final OWLClass startConcept;
    // Here we store all the vertices we want to stay in the graph:
    Set<OmegaTuple> verticesToRetain = new HashSet<>();
    private OmegaSaturator omegaSat;

    public OmegaLoopChecker(OWLClass startConcept, DirectedGraph<OmegaTuple, DefaultEdge> graph,
            OmegaSaturator omegaSat) {
        this.graph = graph;
        this.startConcept = startConcept;
        this.omegaSat = omegaSat;
    }

    /**
     * A tuple t is a root tuple if C_t = A0, A0 in con_t and A0 not in xcon_t.
     */
    private Set<OmegaTuple> getRootTuples() {
        Set<OmegaTuple> result = new HashSet<>();

        for (OmegaTuple tuple : omegaSat.getConceptToOmegaTuple().get(startConcept)) {
            if (tuple.getCon().contains(startConcept)) {
                if (tuple.getXCon() == null) {
                    result.add(tuple);
                } else if (!tuple.getXCon().contains(startConcept)) {
                    result.add(tuple);
                }
            }
        }
        return result;
    }

    /**
     * A tuple t is a root tuple if A0 \in con_t and A0 \not\in xcon_t.
     */
    private Set<OmegaTuple> getRootTuplesOLD() {
        Set<OmegaTuple> result = new HashSet<>();

        for (OmegaTuple tuple : graph.vertexSet()) {
            if (tuple.getCon().contains(startConcept)) {
                if (tuple.getXCon() == null) {
                    result.add(tuple);
                } else if (!tuple.getXCon().contains(startConcept)) {
                    result.add(tuple);
                }
            }
        }
        return result;
    }

    /**
     * Restrict the Omega graph to the component reachable from root tuples if
     * the graph is not looping. Return true if loop-free, otherwise return
     * false. We do graph trimming while checking the looping condition: A tuple
     * t in Omega is looping if there is a path t1...tk through Omega of length
     * at least one s.t., t = t1, con_t = con_tk, and xcon_t = xcon_tk.
     *
     * @return Returns false if we are looping and the reachable component is
     * not built up completely.
     */
    public boolean retainReachableComponentOrFail() {

        // Check for all the root tuples in Omega:
        for (OmegaTuple rootTuple : this.getRootTuples()) {

            // Check for nodes along path with con=con', xcon=xcon'.
            // Start recursion with empty list of visited nodes.
            if (!canDoCompleteDepthFirstTraversal(rootTuple, new HashSet<OmegaTuple>())) {
                //new HashSet<Tuple<Set<OWLClassExpression>>>())) {
                return false;
            }
        }

        // Now we did a complete dfs on the nodes reachable from root tuples.
        // The vertices we visited are stored in verticesToRetain. 
        // Delete the others from graph:
        Set<OmegaTuple> verticesToDelete = new HashSet<>(graph.vertexSet());
        verticesToDelete.removeAll(verticesToRetain);
        graph.removeAllVertices(verticesToDelete);
        return true;
    }

    /**
     * Tries to visit all nodes below currentNode w/o finding duplicate con/xcon
     * pairs. Updates verticesToRetain in the graph. Returns true if possible,
     * false otherwise.
     *
     * @param tuple
     * @param visitedProjectedTuples
     * @return
     */
    private boolean canDoCompleteDepthFirstTraversal(OmegaTuple tuple,
            Set<OmegaTuple> visitedProjectedTuples) {

        // Project current tuple to con and xcon components.
        OmegaTuple projectedTuple = new OmegaTuple(null, null, tuple.getCon(), null, tuple.getXCon());

        // Check whether we already have visited such a tuple of con/xcon:
        if (visitedProjectedTuples.contains(projectedTuple)) {
            // We have an Omega cycle!
            return false;
        }

        // Every tuple that can be visited and has non-duplicating con/xcon
        // stays in the reachable component:
        verticesToRetain.add(tuple);

        visitedProjectedTuples.add(projectedTuple);

        for (DefaultEdge edge : graph.outgoingEdgesOf(tuple)) {
            OmegaTuple successor = graph.getEdgeTarget(edge);
            if (!canDoCompleteDepthFirstTraversal(successor, new HashSet<>(visitedProjectedTuples))) {
                return false;
            }
        }
        return true;
    }

    /**
     * This only checks for Omega loops so far, doesn't do reachability retaining.
     * Used live in OmegaSaturator on incomplete Omega graphs.
     * @return 
     */
    boolean checkOmegaLoopFree() {
        // Check for all the root tuples in Omega:
        for (OmegaTuple rootTuple : this.getRootTuples()) {
            // Check for nodes along path with con=con', xcon=xcon'.
            // Start recursion with empty list of visited nodes.
            if (!canDoCompleteDepthFirstTraversal(rootTuple, new HashSet<OmegaTuple>())) {
                return false;
            }
        }
        return true;
    }
        
}
