/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan.
 * 
 * This file is part of Grind.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind.utilities;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

class Node<T> {

    private final T value;
    private Node<T> parent;
    private final SortedMap<T, Node<T>> children;
    private final Comparator<T> comparator;
    private int minCardinality;

    /**
     * Constructor for root node.
     */
    public Node(Comparator<T> comparator) {
        value = null;
        parent = null;
        children = new TreeMap<>(comparator);
        this.comparator = comparator;
        minCardinality = Integer.MAX_VALUE;
    }

    /**
     * Constructor for non-root nodes.
     *
     * @param value
     * @param parent
     */
    private Node(T value, Node<T> parent) {
        this.value = value;
        this.parent = parent;
        this.comparator = parent.comparator;
        children = new TreeMap<>(parent.comparator);
        minCardinality = Integer.MAX_VALUE;
    }

    public T getValue() {
        return value;
    }

    public int getNumberOfChildren() {
        return children.size();
    }

    public void removeChild(T value) {
        children.remove(value);
    }

    public boolean hasChild(T o) {
        return children.containsKey(o);
    }

    public Node<T> getFirstChild() {
        return children.get(children.firstKey());
    }

    public Node<T> getParent() {
        return parent;
    }

    public Node<T> getChild(T o) {
        return children.get(o);
    }

    public boolean isLeaf() {
        return children.isEmpty();
    }

    public Node<T> addChild(T value, Node<T> parent) {
        Node<T> child = new Node(value, parent);
        children.put(value, child);
        return child;
    }

    public Collection<Node<T>> getChildren() {
        return children.values();
    }

    @Override
    public String toString() {
        String result = children.toString();
//                "( v = " + value + 
//                ", " + children + " )";
        return result;
    }

    public void setParent(Node<T> parent) {
        this.parent = parent;
    }

    public int getMinCardinality() {
        return minCardinality;
    }

    public void setMinCardinality(int minCardinality) {
        this.minCardinality = minCardinality;
    }

};

/**
 *
 * @author seylan
 * @param <T>
 */
public class SetTrie<T> {

    private final Comparator<T> comparator;

    private final Node<T> root;

    /**
     * Construct a new node with 'null' as value for root.
     *
     * @param comparator
     */
    public SetTrie(Comparator<T> comparator) {
        root = new Node<>(comparator);
        this.comparator = comparator;
    }

    /**
     * Answers the query whether this Trie contains a subset of the given set.
     *
     * @param sortedSet
     * @return
     */
    public boolean containsSubsetOf(SortedSet<T> sortedSet) {
        return containsSubsetOf(sortedSet, root);
    }

    private boolean containsSubsetOf(SortedSet<T> sortedSet, Node<T> node) {
        if (node.isLeaf() && (node != root)) {
            return true;
        }
        if (sortedSet.isEmpty()) {
            return false;
        }
        boolean found = false;

        if (node.hasChild(sortedSet.first())) {
            Node<T> nextNode = node.getChild(sortedSet.first());
            SortedSet copy = new TreeSet(((TreeSet) sortedSet).tailSet(sortedSet.first(), false));
            found = containsSubsetOf(copy, nextNode);
        }
        if (!found) {
            SortedSet copy = new TreeSet(((TreeSet) sortedSet).tailSet(sortedSet.first(), false));
            return containsSubsetOf(copy, node);
        } else {
            return true;
        }
    }

    public boolean isEmpty() {
        return root.isLeaf();
    }

    public SortedSet<T> popFirst() {
        SortedSet<T> result = new TreeSet<>(comparator);
        Node<T> node = root;
        while (!node.isLeaf()) {
            node = node.getFirstChild();
            result.add(node.getValue());
        }
        this.removeBranch(node);
        return result;
    }

    public SortedSet<T> popMin() {
        SortedSet<T> result = new TreeSet<>(comparator);
        Node<T> node = root;
        for (Node child : root.getChildren()) {
            if (child.getMinCardinality() == root.getMinCardinality()) {
                node = child;
                result.add(node.getValue());
                break;
            }
        }
        while (!node.isLeaf()) {
            for (Node child : node.getChildren()) {
                if (child.getMinCardinality() == node.getMinCardinality() - 1) {
                    node = child;
                    break;
                }
            }
            result.add(node.getValue());
        }
        this.removeBranch(node);
        return result;
    }

    public void removeSupersets(SortedSet<T> sortedSet) {
        List<Node<T>> nodesToRemove = new ArrayList<>();
        findSupersets(root, new TreeSet<>(sortedSet), nodesToRemove);

        for (Node<T> node : nodesToRemove) {
            removeBranch(node);
        }
    }

    public boolean insert(SortedSet<T> sortedSet) {
        if (containsSubsetOf(sortedSet)) {
            return false;
        }
        List<Node<T>> nodesToRemove = new ArrayList<>();
        findSupersets(root, new TreeSet<>(sortedSet), nodesToRemove);

        for (Node<T> node : nodesToRemove) {
            removeBranch(node);
        }
        insertNonExisting(sortedSet);
        return true;
    }

    public Set<Set<T>> toSetOfSets() {
        Set<Set<T>> result = new HashSet<>();
        Set<T> currentSet = new HashSet<>();
        constructSetOfSets(root, currentSet, result);
        return result;
    }

    private void constructSetOfSets(Node<T> node, Set<T> currentSet, Set<Set<T>> result) {
        if (node.isLeaf() && node != root) {
            result.add(currentSet);
            return;
        }
        for (Node<T> child : node.getChildren()) {
            Set<T> newCurrent = new HashSet<>(currentSet);
            newCurrent.add(child.getValue());
            constructSetOfSets(child, newCurrent, result);
        }
    }

    private void insertNonExisting(SortedSet<T> sortedSet) {
        int newHeight = sortedSet.size();
        Node<T> currentNode = root;
        if (!sortedSet.isEmpty() && root.getMinCardinality() > newHeight) {
            root.setMinCardinality(newHeight);
        }
        for (T element : sortedSet) {
            if (currentNode.hasChild(element)) {
                currentNode = currentNode.getChild(element);
            } else {
                currentNode = currentNode.addChild(element, currentNode);
            }
            if (currentNode.getMinCardinality() > newHeight) {
                currentNode.setMinCardinality(newHeight);
            }
            newHeight--;
        }
    }

    private void arrangeHeight(Node<T> node) {
        if (node == null) {
            return;
        }
        int minChildCardinality = Integer.MAX_VALUE;
        for (Node<T> child : node.getChildren()) {
            if (child.getMinCardinality() < minChildCardinality) {
                minChildCardinality = child.getMinCardinality();
            }
        }
        // root gets the same height as its longest child
        if (node == root) {
            node.setMinCardinality(minChildCardinality);
        } else {
            node.setMinCardinality(minChildCardinality + 1);
        }
        arrangeHeight(node.getParent());
    }

    @Override
    public String toString() {
        return root.toString();
    }

    private void removeBranch(Node<T> node) {
        if (node == root) {
            return;
        }
        Node<T> par = node.getParent();
        while ((par.getNumberOfChildren() == 1) && (par != root)) {
            node = par;
            par = node.getParent();
        }
        node.setParent(null);
        par.removeChild(node.getValue());
        arrangeHeight(par);
    }

    public int getMinCardinality() {
        return root.getMinCardinality();
    }

    private void findSupersets(Node<T> node, TreeSet<T> sortedSet, List<Node<T>> nodesToRemove) {
        if (sortedSet.isEmpty()) {
            nodesToRemove.add(node);
            return;
        }
        if (node.isLeaf()) {
            return;
        }

        T value = sortedSet.first();

        for (Node<T> child : node.getChildren()) {
            if (comparator.compare(child.getValue(), value) < 0) {
                findSupersets(child, new TreeSet<>(sortedSet), nodesToRemove);
            } else if (comparator.compare(child.getValue(), value) == 0) {
                findSupersets(child, new TreeSet<>(sortedSet.tailSet(value, false)), nodesToRemove);
            } else {
                break;
            }
        }
    }

}
