/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.context;

import com.google.common.collect.Lists;
import de.unibremen.informatik.tdki.grind.context.OriginalOntology;
import de.unibremen.informatik.tdki.grind2.cq.CQException;
import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import de.unibremen.informatik.tdki.grind2.utilities.Signature;
import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLObjectPropertyExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.model.PrefixManager;
import org.semanticweb.owlapi.util.DefaultPrefixManager;

/**
 *
 * @author hansen
 */
public class CQContext {

    private static CQContext INSTANCE = new CQContext();

    private final boolean DEBUG = false;

    private boolean verbose = false;
    
    private int timeout;

    private boolean queryfile;

    private boolean loaded = false;
    private OWLOntology fromFileOnt;
    private OriginalOntology originalOnt;

    private final PrefixManager pm;

    private final Map<OWLClassExpression, Variable> clsAnnotationMap;
    private final Map<OWLObjectProperty, Variable> roleAnnotationMap;
    private final Map<OWLClass, OWLObjectSomeValuesFrom> spurious2ExistMap;
    private final Map<List, OWLClass> rCx2SpuriousMap;

    private final Map<OWLClass, Integer> annotatedClasses2VarsMap;
    private final Map<OWLClass, OWLClassExpression> annotatedClasses2ClassMap;
    private final Map<OWLObjectProperty, Integer> annotatedRoles2VarsMap;
    private final Map<OWLObjectProperty, OWLObjectProperty> annotatedRoles2RoleMap;

    private int counter = 0;

    public CQContext() {
//        pm = new DefaultPrefixManager("http://uni-bremen.de/informatik/tdki#");
//        pm = new DefaultPrefixManager("http://purl.obolibrary.org/obo/");
        pm = new DefaultPrefixManager("#");
        clsAnnotationMap = new HashMap<>();
        roleAnnotationMap = new HashMap<>();
        spurious2ExistMap = new HashMap<>();
        rCx2SpuriousMap = new HashMap<>();
        annotatedClasses2ClassMap = new HashMap<>();
        annotatedClasses2VarsMap = new HashMap<>();
        annotatedRoles2RoleMap = new HashMap<>();
        annotatedRoles2VarsMap = new HashMap<>();
    }

    public static CQContext getINSTANCE() {
        return INSTANCE;
    }

    public int getTimeout() {
        return timeout;
    }

    public void setTimeout(int timeout) {
        this.timeout = timeout;
    }

    public static void refresh() {
        // Save values of verbose/norewrite:
        boolean verbose = getINSTANCE().isVerbose();

        INSTANCE = new CQContext();
        getINSTANCE().setVerbose(verbose);
    }

    public boolean isVerbose() {
        return verbose;
    }

    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    public boolean isQueryfile() {
        return queryfile;
    }

    public void setQueryfile(boolean queryfile) {
        this.queryfile = queryfile;
    }

    public boolean isDEBUG() {
        return DEBUG;
    }

    public int getCounter() {
        return counter++;
    }

    public PrefixManager getPm() {
        return pm;
    }

    @Deprecated
    public void loadFromFile(String filename) throws OWLOntologyCreationException {
        OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
        File file = new File(filename);
        fromFileOnt = manager.loadOntologyFromOntologyDocument(file);
        this.originalOnt = new OriginalOntology(fromFileOnt, manager);
        loaded = true;
    }

    public OriginalOntology getOriginalOnt() {
        return originalOnt;
    }

    public Signature getOriginalSignature() throws CQException {
        if (!loaded) {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried to getOrigSignature without prior loading of ontology!");
        }
        Set<OWLClass> setOfClasses = originalOnt.getConceptNames();
        //fromFileOnt.getClassesInSignature();
        Set<OWLObjectProperty> setOfRoles = originalOnt.getRoleNames();
        //fromFileOnt.getObjectPropertiesInSignature();
        return new Signature(setOfClasses, setOfRoles);
    }

    public OWLOntology getFromFileOntology() throws CQException {
        if (!loaded) {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried to getOnt without prior loading of ontology!");
        }
        return fromFileOnt;
    }

    private void addSpurious(OWLClass spurious, OWLObjectSomeValuesFrom existRG) {
        this.spurious2ExistMap.put(spurious, existRG);
    }

    public boolean hasSpurious(OWLClass spurious) {
        return spurious2ExistMap.containsKey(spurious);
    }

    public OWLObjectSomeValuesFrom getExistRCFromSpurious(OWLClass spurious) throws CQException {
        if (this.spurious2ExistMap.containsKey(spurious)) {
            return this.spurious2ExistMap.get(spurious);
        } else {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried get existsRC from spurious that is not in map!");
        }
    }

    /** Annotate a concept name [A] as [A^x].
     *
     * @param cls
     * @param x
     * @return 
     */
    public OWLClass annotateCls(OWLClass cls, Variable x) {
        String newClsString = cls.getIRI().getFragment()
                + "^" + x.toString();        
        OWLClass newCls = OWLManager.getOWLDataFactory().
                getOWLClass(newClsString, CQContext.getINSTANCE().getPm());
        CQContext.getINSTANCE().addAnnotation(cls, newCls, x);
        return newCls;
    }

    /**
     * Annotate a subconcept [exist R.C] as [exist R^x.C].
     *
     * @param existRG
     * @param x
     * @return
     * @throws CQException
     */
    public OWLObjectSomeValuesFrom annotateExist(OWLObjectSomeValuesFrom existRG, Variable x) throws CQException {
        OWLObjectPropertyExpression role = existRG.getProperty();
        if (!(role instanceof OWLObjectProperty)) {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Found " + role + ", a non-simple role!");
        }
        OWLObjectProperty simpleRole = (OWLObjectProperty) role;
        OWLClassExpression g = existRG.getFiller();
        String newRoleString = simpleRole.getIRI().getFragment() + "^" + x.toString();
        
        OWLObjectProperty newRole = OWLManager.getOWLDataFactory()
                .getOWLObjectProperty(newRoleString, CQContext.getINSTANCE().getPm());

        CQContext.getINSTANCE().addAnnotation(simpleRole, newRole, x);
        OWLObjectSomeValuesFrom newExistRxG = OWLManager.getOWLDataFactory()
                .getOWLObjectSomeValuesFrom(newRole, g);
        return newExistRxG;
    }

    /**
     * Introduce new spurious concept name [A_exRC^x] for a subconcept [exist
     * R.C]. Detect if we already encountered R and C!
     *
     * @param existRG
     * @param x
     * @return
     * @throws CQException
     */
    public OWLClass annotateSpurious(OWLObjectSomeValuesFrom existRG, Variable x) throws CQException {
        OWLObjectPropertyExpression role = (OWLObjectPropertyExpression) existRG.getProperty();
        if (!(role instanceof OWLObjectProperty)) {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Found " + role + ", a non-simple role!");
        }
        OWLObjectProperty simpleRole = (OWLObjectProperty) role;

        // Construct <r, C, x> and check whether this tuple is already known:
        List rCxTuple = Lists.newArrayList(simpleRole, existRG.getFiller(), x);
        if (rCx2SpuriousMap.containsKey(rCxTuple)) {
            return rCx2SpuriousMap.get(rCxTuple);
        }

        // Unknown. Construct new spurious concept name:
        OWLClass newSpuriousCls = OWLManager.getOWLDataFactory().
                getOWLClass("A_exRC" + CQContext.getINSTANCE().getCounter()
                        + "^" + x.toString(), CQContext.getINSTANCE().getPm());
        CQContext.getINSTANCE().addAnnotation(existRG, newSpuriousCls, x);
        CQContext.getINSTANCE().addSpurious(newSpuriousCls, existRG);

        // Make it known:
        rCx2SpuriousMap.put(rCxTuple, newSpuriousCls);
        return newSpuriousCls;
    }

    private void addAnnotation(OWLClassExpression cls, OWLClass newCls, Variable x) {
        this.clsAnnotationMap.put(cls, x);
        this.annotatedClasses2ClassMap.put(newCls, cls);
        if (x.getIndices().size() != 1) {
            System.err.println("[" + this.getClass().getSimpleName()
                    + "] Tried to annotate " + cls + " with variable " + x + ", which has several indices!");
            System.exit(-1);
        }
        int index = x.getIndices().iterator().next();
        this.annotatedClasses2VarsMap.put(newCls, index);
    }

    private void addAnnotation(OWLObjectProperty role, OWLObjectProperty newRole, Variable x) {
        this.roleAnnotationMap.put(role, x);
        this.annotatedRoles2RoleMap.put(newRole, role);
        if (x.getIndices().size() != 1) {
            System.err.println("[" + this.getClass().getSimpleName()
                    + "] Tried to annotate " + role + " with variable " + x + ", which has several indices!");
            System.exit(-1);
        }
        int index = x.getIndices().iterator().next();
        this.annotatedRoles2VarsMap.put(newRole, index);
    }

    public boolean hasAnnotation(OWLClass cls) {
        return this.clsAnnotationMap.containsKey(cls);
    }

    public boolean hasAnnotation(OWLObjectProperty role) {
        return this.roleAnnotationMap.containsKey(role);
    }

    public Variable getAnnotation(OWLClass cls) throws CQException {
        if (this.clsAnnotationMap.containsKey(cls)) {
            return this.clsAnnotationMap.get(cls);
        } else {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried get annotation that is not in map!");
        }
    }

    public Variable getAnnotation(OWLObjectProperty role) throws CQException {
        if (this.roleAnnotationMap.containsKey(role)) {
            return this.roleAnnotationMap.get(role);
        } else {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried get annotation that is not in map!");
        }
    }

    public Map<OWLClass, Integer> getAnnotatedClasses2VarsMap() {
        return annotatedClasses2VarsMap;
    }

    public Map<OWLClass, OWLClassExpression> getAnnotatedClasses2ClassMap() {
        return annotatedClasses2ClassMap;
    }

    public Map<OWLObjectProperty, Integer> getAnnotatedRoles2VarsMap() {
        return annotatedRoles2VarsMap;
    }

    public Map<OWLObjectProperty, OWLObjectProperty> getAnnotatedRoles2RoleMap() {
        return annotatedRoles2RoleMap;
    }

    public OWLClass getClsN() {
        return OWLManager.getOWLDataFactory().getOWLClass("FreshConceptN", CQContext.getINSTANCE().getPm());
    }

    @Deprecated
    public static void load(String filename) throws OWLOntologyCreationException {
        long t = System.currentTimeMillis();
        CQContext.refresh();
        CQContext.getINSTANCE().loadFromFile(filename);
        if (CQContext.getINSTANCE().isVerbose()) {
            System.out.println("### ");
            System.out.println("### Loading of ontology " + filename);
            System.out.println("### took " + (System.currentTimeMillis() - t) + " ms.");
            System.out.println("### ");
        }
    }

}
