/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.context;

import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind2.cq.CQ;
import de.unibremen.informatik.tdki.grind2.cq.CQException;
import de.unibremen.informatik.tdki.grind2.cq.Partition;
import de.unibremen.informatik.tdki.grind2.cq.RootedCQ;
import de.unibremen.informatik.tdki.grind2.cq.Splitting;
import de.unibremen.informatik.tdki.grind2.cq.TreeCQ;
import de.unibremen.informatik.tdki.grind2.cq.components.Atom;
import de.unibremen.informatik.tdki.grind2.cq.components.ClassAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.RoleAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import de.unibremen.informatik.tdki.grind2.syntax.Annotations;
import de.unibremen.informatik.tdki.grind2.utilities.OntologyUtils;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.AxiomType;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;

/**
 *
 * @author hansen
 */
public class RootedReasoner {

    public static Set<OWLAxiom> getOntologyTpForP(Set<OWLAxiom> axioms, RootedCQ forkrewrP) throws OWLOntologyCreationException, CQException {
        // Convert to Set of OWLSubClassAxioms:
        Set<OWLSubClassOfAxiom> setOfOrigCIs = new HashSet<>();        
        for (OWLAxiom ax : axioms) {
            if (ax instanceof OWLSubClassOfAxiom) {
                OWLSubClassOfAxiom scax = (OWLSubClassOfAxiom) ax;
                setOfOrigCIs.add(scax);
            }
        }
        // Copy it:
        Set<OWLSubClassOfAxiom> setOfNewCIs = new HashSet<>(setOfOrigCIs);

        // For every old CI:
        for (OWLSubClassOfAxiom ci : setOfOrigCIs) {

            OWLClassExpression lhs = ci.getSubClass();
            OWLClassExpression rhs = ci.getSuperClass();

            // Generate C^x_R => D^x_R (both sides only _R!)
            // Work on tlcs:
            Set<OWLClassExpression> oldLhsTlcs = lhs.asConjunctSet();
            Set<OWLClassExpression> oldRhsTlcs = rhs.asConjunctSet();

            for (Variable x : forkrewrP.getVars()) {
                Set<OWLClassExpression> newLhsTlcs = Annotations.rhsAnnotation(oldLhsTlcs, x);
                Set<OWLClassExpression> newRhsTlcs = Annotations.rhsAnnotation(oldRhsTlcs, x);

                OWLClassExpression newRhs = OntologyUtils.getClsOrIntersection(newRhsTlcs);
                OWLClassExpression newLhs = OntologyUtils.getClsOrIntersection(newLhsTlcs);
                OWLSubClassOfAxiom newCI = OWLManager.getOWLDataFactory()
                        .getOWLSubClassOfAxiom(newLhs, newRhs);
                setOfNewCIs.add(newCI);
            }
        }

        // Construct (forall x roots of C) C^x_R  => N:
        OWLClass clsN = CQContext.getINSTANCE().getClsN();

        // In the definition of this rule, we need the graph G^-, i.e. the query
        // without core roles.
        Set<Atom> withoutCoreRolesAtomSet = new HashSet<>(forkrewrP.getSetOfAtoms());
        CQ core = forkrewrP.getCore();
        Set<Atom> coreAtomSet = core.getSetOfAtoms();
//        System.out.println("Core: " + core.getSetOfAtoms());
        for (Atom at : coreAtomSet) {
            if (at instanceof RoleAtom) {
                withoutCoreRolesAtomSet.remove(at);
            }
        }
        CQ forkrewrPWithoutCoreRoles = new CQ(forkrewrP.getAvars(), withoutCoreRolesAtomSet);
//        System.out.println("forkrewrPWithoutCoreRoles: " + forkrewrPWithoutCoreRoles);

        // Now we have to look for the concepts (trees) C that hang at every
        // core var x: Add _R'd version to the big sqcap!
        Set<OWLClassExpression> setOfXAnnotatedConcepts = new HashSet<>();
        for (Variable x : core.getVars()) {
            TreeCQ tree = forkrewrPWithoutCoreRoles.getTreeCQInAcyclicPartAt(x);
            OWLClassExpression conceptAtX = tree.getConcept();
//            System.out.println("Concept found at " + x + " is: " + conceptAtX);
            Set<OWLClassExpression> setOfXConcept = Sets.newHashSet(conceptAtX.asConjunctSet());
            Set<OWLClassExpression> setOfAnnotatedConceptAtX = Annotations.rhsAnnotation(setOfXConcept, x);
            setOfXAnnotatedConcepts.addAll(setOfAnnotatedConceptAtX);
        }

        OWLClassExpression allAnnClasses = OntologyUtils.getClsOrIntersection(setOfXAnnotatedConcepts);
        OWLSubClassOfAxiom lastCI = OWLManager.getOWLDataFactory()
                .getOWLSubClassOfAxiom(allAnnClasses, clsN);
        setOfNewCIs.add(lastCI);

        // Have to generate new OWLAxiom set. Why? 
        // Why OWLSubClassAxiom not subclass of OWLAxiom??
        Set<OWLAxiom> newCIsAsAxioms = new HashSet<>();
        for (OWLSubClassOfAxiom ci : setOfNewCIs) {
            newCIsAsAxioms.add(ci);
        }
        return newCIsAsAxioms;
    }

    public static OWLOntology getOntologyTpminFromTpAndSplittings(OWLOntology ontTp, Set<Splitting> allSplittingsOfQtoP) throws OWLOntologyCreationException, CQException {

        // TPmin := Tp...
        Set<OWLSubClassOfAxiom> setOfNewCIs = new HashSet<>(ontTp.getAxioms(AxiomType.SUBCLASS_OF));
        // ... extended by one CI for each Splitting:

        for (Splitting splitting : allSplittingsOfQtoP) {
            Partition part = splitting.getPartition();
            Map<Variable, Variable> nu = splitting.getNu();

            Set<OWLClassExpression> lhsConjuncts = new HashSet<>();

            // (1) CNs at R variables:
            Set<Atom> rPartAtoms = part.getrPart().getSetOfAtoms();
            for (Atom at : rPartAtoms) {
                if (at instanceof ClassAtom) {
                    ClassAtom cat = (ClassAtom) at;
                    OWLClass cls = cat.getCls();
                    Variable x = cat.getVar();
                    Variable nuOfx = nu.get(x);
                    OWLClass annotatedCls = CQContext.getINSTANCE().annotateCls(cls, nuOfx);
                    lhsConjuncts.add(annotatedCls);
                }
            }

            // (2) Existential trees that hang at tree roots:
            for (int i = 0; i < part.getTreeRootParents().size(); i++) {
                Variable treeRootParent = part.getTreeRootParents().get(i);
                Variable treeRoot = part.getTreeRoots().get(i);
                TreeCQ tree = part.getTrees().get(i);
                OWLObjectProperty role = part.getTreeRootRoles().get(i);
                OWLClassExpression treeConcept = tree.getConcept();
                Variable nuOfParent = nu.get(treeRootParent);

                OWLObjectSomeValuesFrom existRTreeConcept
                        = OWLManager.createOWLOntologyManager().getOWLDataFactory().getOWLObjectSomeValuesFrom(role, treeConcept);
                OWLObjectSomeValuesFrom annotatedExistRTreeConcept
                        = CQContext.getINSTANCE().annotateExist(existRTreeConcept, nuOfParent);

                lhsConjuncts.add(annotatedExistRTreeConcept);
            }

            // Finally build the CI from (1) and (2) conjuncts:
            OWLClassExpression lhsConjunction = OWLManager.getOWLDataFactory().getOWLObjectIntersectionOf(lhsConjuncts);
            OWLClass clsN = CQContext.getINSTANCE().getClsN();
            OWLSubClassOfAxiom lhsConjunctsImplyN = OWLManager.getOWLDataFactory().getOWLSubClassOfAxiom(lhsConjunction, clsN);
            setOfNewCIs.add(lhsConjunctsImplyN);
        }

        // Have to generate new OWLAxiom set. Why? 
        // Why OWLSubClassAxiom not subclass of OWLAxiom??
        Set<OWLAxiom> newCIsAsAxioms = new HashSet<>();
        for (OWLSubClassOfAxiom ci : setOfNewCIs) {
            newCIsAsAxioms.add(ci);
        }
        return OWLManager.createOWLOntologyManager().createOntology(newCIsAsAxioms);
    }

}
