/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.control;

import de.unibremen.informatik.tdki.grind.context.Context;
import de.unibremen.informatik.tdki.grind.control.GammaAndOmegaCaller;
import de.unibremen.informatik.tdki.grind.datalog.GammaToDatalogRewriter;
import de.unibremen.informatik.tdki.grind.datalog.OmegaToDatalogRewriter;
import de.unibremen.informatik.tdki.grind.gammaconstruction.Gamma;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaSaturator;
import de.unibremen.informatik.tdki.grind.gammaconstruction.GammaTuple;
import de.unibremen.informatik.tdki.grind.omegaconstruction.Omega;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaLoopChecker;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaSaturator;
import de.unibremen.informatik.tdki.grind.omegaconstruction.OmegaTuple;
import de.unibremen.informatik.tdki.grind.statistics.DetailsLogEntry;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.context.RootedReasoner;
import de.unibremen.informatik.tdki.grind2.cq.Splitting;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogProgram;
import java.io.File;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.DefaultEdge;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectSomeValuesFrom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyManager;

/**
 *
 * @author hansen
 */
public class TpGammaTpminOmegaCaller {

    private static final int S = 1000; // 1 s = 1000 ms
    private static final int MIN = 60_000;
    private static final int H = 60 * 60_000;

//    private static final int TIMEOUT = 30 * MIN;
//    private static int GAMMA_TIMEOUT = 15 * S;
//    private static int OMEGA_TIMEOUT = 15 * S;

    private static final boolean DEBUG = false;

    private final OWLOntologyManager manager;
    private File file;
    private final OWLOntology ontTp;
//    private OWLOntology ontTpmin;
    private Context contextTp;
    private Context contextTpmin;
    private long tStartOfThisConcept;
    private final long t0;
    private PrintWriter output;
    private int numberOfGammaCycles;
    private int numberOfGammaCycleFOno, numberOfGammaCycleFOyes;
    private int numberOfGammaTimeouts;
    private int numberOfOmegaTimeouts;
    private final boolean verbose;
    private final Set<Splitting> allSplittingsForAllFRs;

    private DatalogProgram resultingProgram;

    public TpGammaTpminOmegaCaller(OWLOntology ontTp, Set<Splitting> allSplittingsForAllFRs) throws Exception {
        // Start global time count:
        t0 = System.currentTimeMillis();

        this.manager = OWLManager.createOWLOntologyManager();
        this.ontTp = ontTp;
        this.verbose = CQContext.getINSTANCE().isVerbose();
        this.allSplittingsForAllFRs = allSplittingsForAllFRs;

        //overviewLogEntry = new OverviewLogEntry(CQContext.getInstance().getFilename());
        openOntologyTp(ontTp);
//        openOntologyTpmin(ontTpmin);
    }

    public DatalogProgram getResultingProgram() {
        return resultingProgram;
    }

    /**
     * Grind2 addition, run Gamma and Omega with predefined OWLClass instance.
     *
     * @param startconcept
     * @return
     * @throws Exception
     */
    public int runSingleConceptname(OWLClass startconcept) throws Exception {

        tStartOfThisConcept = System.currentTimeMillis();
        DetailsLogEntry detailsLogEntry = new DetailsLogEntry();
        int retval = runGammaAndOmega(startconcept, 1, detailsLogEntry);
        if (verbose) {
            System.out.println("\nTotal time: " + (System.currentTimeMillis() - tStartOfThisConcept) + " ms.\n");
        }
        detailsLogEntry.setTimeSpent(System.currentTimeMillis() - tStartOfThisConcept);
        return retval;
    }

    private void openOntologyTp(OWLOntology ontTp) {

        contextTp = new Context();
        contextTp.setOntAndManager(ontTp, manager);
        manager.removeOntology(ontTp);

        if (verbose) {
            System.out.println("  Normalising the ontology Tp: " + (System.currentTimeMillis() - t0) + " ms.");
            System.out.println("\n  Size of sub(Tp):             " + contextTp.getOriginalOnt().getSubconcepts().size());
        }
        //overviewLogEntry.setNoOfConceptnames(1);
        //overviewLogEntry.setNoOfRolenames(contextTp.getOriginalOnt().getRoleNames().size());
    }

    private void openOntologyTpmin(OWLOntology ontTpmin) {

        contextTpmin = new Context();
        contextTpmin.setOntAndManager(ontTpmin, manager);
        manager.removeOntology(ontTpmin);

        if (verbose) {
            System.out.println("  Normalising the ontology Tpmin: " + (System.currentTimeMillis() - t0) + " ms.");
            System.out.println("\n  Size of sub(Tpmin):             " + contextTpmin.getOriginalOnt().getSubconcepts().size());
        }
        //overviewLogEntry.setNoOfConceptnames(1);
        //overviewLogEntry.setNoOfRolenames(contextTpmin.getOriginalOnt().getRoleNames().size());
    }

    private int runGammaAndOmega(OWLClass startConcept, int currentClass, DetailsLogEntry detailsLogEntry) throws Exception {

        boolean ontTpminAlreadyCreated = false;

        detailsLogEntry.setConceptName(startConcept);

        if (verbose) {
            System.out.print("(" + currentClass + " / 1) ");
            System.out.println("Working on class " + startConcept.getIRI().getFragment());
        }

        // * Saturate Gamma:
        GammaSaturator gammaSat = new GammaSaturator(startConcept, detailsLogEntry, contextTp);
        long usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
        gammaSat.saturate(CQContext.getINSTANCE().getTimeout() - usedUntilNow);

        // Main loop, do Gamma - Omega - Gamma - Omega...
        while (true) {

            // * if Gamma timeout:
            if (detailsLogEntry.isGammaTimeout()) {
                numberOfGammaTimeouts++;
                return -1;
            }

            // * if Gamma acyclic:
            Gamma gamma = gammaSat.getGamma();
            gamma.createGammaGraph();

            if (!gamma.isCyclic()) {

                Set<GammaTuple> gammaWithoutSpurious = new HashSet<>();
                outer:
                for (GammaTuple tuple : gamma.getGammaTuples()) {
                    if (tuple.getConcept().toString().contains("A_exRC")) {
                        continue;
                    }
                    for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                        if (concept.toString().contains("A_exRC")) {
                            continue outer;
                        }
                    }
                    gammaWithoutSpurious.add(tuple);
                }

                if (verbose) {
                    System.out.println("Concept " + startConcept.getIRI().getFragment()
                            + " is FO-rewritable (Gamma acyclic, size " + gammaWithoutSpurious.size() + ").\n");
                }
                GammaToDatalogRewriter dr = new GammaToDatalogRewriter(gamma, startConcept, contextTp);
                //System.out.println(dr.gammaToDatalogStrings());
                this.resultingProgram = dr.gammaToDatalogProgram();

                return 0;
            }

            // Increase number of Gamma cycles only once:
            if (!detailsLogEntry.isCyclicGamma()) {
                detailsLogEntry.setCyclicGamma(true);
                numberOfGammaCycles++;
            }

            detailsLogEntry.setNoOfGammaTuplesOnCycle(gamma.numberOfTuplesOnCycle());

            if (verbose) {
                System.out.println("Gamma graph:");
                System.out.println("  Number of vertices:     " + gamma.getGammaGraph().vertexSet().size());
                System.out.println("  Number of edges:        " + gamma.getGammaGraph().edgeSet().size());
                //System.out.print("Distr. of ex. restr. in S: ");
                //System.out.println("    " + countExistentials(gamma.getGammaTuples()));
                System.out.println("  # of vert. lying on cycle: " + gamma.numberOfTuplesOnCycle());
                System.out.println("  distr. of their ex. restr.:");
                System.out.println("    " + countExistentials(gamma.getTuplesOnCycle()));
                //countSimpleCycles(graph);  // VERY costly!!
                System.out.println();
            }
            if (DEBUG) {
                System.out.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                System.out.println("GAMMA:\n");
                System.out.println(GammaSaturator.shortenIDs(gamma.toString()));
                System.out.println("\n-------\n\n");
            }

            // * Saturate Omega:
            OWLOntology ontTpmin = RootedReasoner.getOntologyTpminFromTpAndSplittings(ontTp, allSplittingsForAllFRs);
            openOntologyTpmin(ontTpmin);
            OmegaSaturator omegaSat = new OmegaSaturator(gamma, detailsLogEntry, startConcept, contextTpmin);
            usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
            try {
                omegaSat.saturate(CQContext.getINSTANCE().getTimeout() - usedUntilNow);
            } catch (OutOfMemoryError e) {
                System.err.println(e);
                omegaSat.dumpInternals();
                System.exit(-1);
            }
            Omega omega = omegaSat.getOmega();

            // * if Omega timeout:
            if (detailsLogEntry.isOmegaTimeout()) {
                numberOfOmegaTimeouts++;
                return -2;
            }

            detailsLogEntry.setNoOfOmegaTuples(omega.getOmegaTuples().size());

            // * if Omega cyclic:
            DirectedGraph<OmegaTuple, DefaultEdge> omegaGraph = omegaSat.getOmegaGraph();
            OmegaLoopChecker omegaLoopChecker = new OmegaLoopChecker(startConcept, omegaGraph, omegaSat);
            // To construct the Datalog rewriting, we only need the component 
            // of the graph that is reachable from root tuples. Note that 
            // we actually do the check for Omega loops here and get the 
            // reachable component for free.
            boolean omegaLoopFree = omegaLoopChecker.retainReachableComponentOrFail();
            if (!omegaLoopFree) {
                if (DEBUG) {
                    System.out.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                    System.out.println("OMEGA\n");
                    System.out.println(GammaSaturator.shortenIDs(omega.toString()));
                    System.out.println("\n-------\n\n");
                }
                if (CQContext.getINSTANCE().isVerbose()) {
                    System.out.println("Concept " + startConcept.getIRI().getFragment()
                            + " is not FO-rewritable (Omega cyclic).");
                }
                numberOfGammaCycleFOno++;
                return 1;
            }

            // * if Omega acyclic and Gamma not premature:
            if (omegaLoopFree && !gammaSat.isPrematureReturnBecauseOfCycle()) {
                if (DEBUG) {
                    System.out.print("[DEBUG " + this.getClass().getSimpleName() + "] ");
                    System.out.println("OMEGA\n");
                    System.out.println(GammaSaturator.shortenIDs(omega.toString()));
                    System.out.println("\n-------\n\n");
                }

                if (verbose) {
                    System.out.println("Concept " + startConcept.getIRI().getFragment() + " is FO-rewritable (Omega cancelling).\n");
                }
                OmegaToDatalogRewriter dr = new OmegaToDatalogRewriter(omegaSat, startConcept, contextTpmin);
                //System.out.println(dr.omegaToDatalogStrings());
                this.resultingProgram = dr.omegaToDatalogProgram();

                numberOfGammaCycleFOyes++;
                return 0;
            }

            // * Resume Gamma saturation:
            usedUntilNow = System.currentTimeMillis() - tStartOfThisConcept;
            gammaSat.resumeSaturation(CQContext.getINSTANCE().getTimeout() - usedUntilNow);

        } // while true
    }

    /**
     * This gets all Gamma tuples that lie on a cycle.
     *
     * @param setOfTuplesOnCycle
     * @return
     */
    public static Map<Integer, Integer> countExistentials(Set<GammaTuple> setOfTuplesOnCycle) {

        Map<Integer, Integer> stat = new HashMap<>();

        for (GammaTuple tuple : setOfTuplesOnCycle) {

//            System.err.print("[DEBUG GammaAndOmegaCaller] ");
//            System.err.println("Gamma tuple on cycle: " + tuple);
            int numberOfExistentials = 0;
            for (OWLClassExpression concept : tuple.getSetOfConcepts()) {
                if (concept instanceof OWLObjectSomeValuesFrom) {
                    numberOfExistentials++;
                }
            }
            if (!stat.containsKey(numberOfExistentials)) {
                stat.put(numberOfExistentials, 1);
            } else {
                stat.put(numberOfExistentials, stat.get(numberOfExistentials) + 1);
            }
        }
        return stat;
    }
}
