/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.control.rewriters;

import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.context.RootedReasoner;
import de.unibremen.informatik.tdki.grind2.control.TpGammaTpminOmegaCaller;
import de.unibremen.informatik.tdki.grind2.cq.CQ;
import de.unibremen.informatik.tdki.grind2.cq.CQException;
import de.unibremen.informatik.tdki.grind2.cq.Partition;
import de.unibremen.informatik.tdki.grind2.cq.RootedCQ;
import de.unibremen.informatik.tdki.grind2.cq.Splitting;
import de.unibremen.informatik.tdki.grind2.cq.TqCQ;
import de.unibremen.informatik.tdki.grind2.cq.components.Atom;
import de.unibremen.informatik.tdki.grind2.cq.components.RoleAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogProgram;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogRule;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogQueryAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogRoleAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogVariable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;

/**
 *
 * @author hansen
 */
public class RootedCQRewriter extends QueryRewriter {

    private int numberOfFRs;
    private int numberOfSplittings;

    public RootedCQRewriter(Set<OWLAxiom> axioms) {
        super(axioms);
    }

    public int getNumberOfFRs() {
        return numberOfFRs;
    }

    public int getNumberOfSplittings() {
        return numberOfSplittings;
    }

    @Override
    public int rewrite(CQ qOrig) throws CQException, OWLOntologyCreationException, Exception {

        RootedCQ q0 = qOrig.asRootedCQ();
        Set<DatalogVariable> avarsAsDatalogVariables = new HashSet<>();
        for (Variable x : q0.getAvars()) {
            avarsAsDatalogVariables.add(new DatalogVariable(x.getSingleIndex()));
        }

        Set<RootedCQ> allForkrewrOfQ0 = q0.getForkRewritings();
        this.numberOfFRs = allForkrewrOfQ0.size();

        //
        // Start the Optimization first!
        //
        Map<RootedCQ, DatalogProgram> forkrewr2DlogMap = new HashMap<>();
        Map<RootedCQ, Set<DatalogRoleAtom>> forkrewr2dlogRoleAtoms = new HashMap<>();
        Map<RootedCQ, Set<DatalogVariable>> forkrewr2dlogVariables = new HashMap<>();

        boolean foundOneBadForkrewriting = false;
        outer:
        for (RootedCQ p : allForkrewrOfQ0) {

            if (CQContext.getINSTANCE().isVerbose()) {
                System.out.println("fork rewriting p: " + p);
            }
            CQ coreOfP = p.getCore();
            Set<Atom> coreAtoms = coreOfP.getSetOfAtoms();
            Set<Atom> coreRoleAtoms = new HashSet<>();
            for (Atom at : coreAtoms) {
                if (at instanceof RoleAtom) {
                    RoleAtom rat = (RoleAtom) at;
                    coreRoleAtoms.add(rat);
                }
            }
            Set<Atom> setOfPAtoms = p.getSetOfAtoms();
            
            // Question: Do we construct pAsTqCQ WITHOUT core role atoms??
            Set<Atom> pWithoutCoreRoleAtoms = Sets.difference(setOfPAtoms, coreRoleAtoms);

            Set<Variable> coreVars = coreOfP.getVars();
            
            // Special case: core might be empty if whole forkrewr is tree:
            if (coreVars.isEmpty()){
                for (Variable x:p.getVars()) {
                    // Check if it's the (unique, maybe identified!) avar:
                    if (x.containsSomeAvarOf(p.getAvars())) {
                        coreVars.add(x);
                    }
                }
            }

            //TqCQ pAsTqCQ = new TqCQ(coreVars, pWithoutCoreRoleAtoms);
            TqCQ pAsTqCQ = new TqCQ(coreVars, setOfPAtoms);
            
            forkrewr2dlogRoleAtoms.put(p, pAsTqCQ.getCoreDatalogRoleAtoms());
            forkrewr2dlogVariables.put(p, pAsTqCQ.getCoreDatalogVariables());

            TqCQRewriter tqCQrewr = new TqCQRewriter(this.getAxioms());

            int optiRetval = tqCQrewr.rewrite(pAsTqCQ);

            switch (optiRetval) {
                case 1: {
                    if (CQContext.getINSTANCE().isVerbose()) {
                        System.out.println("rCQ Optimization found nonFOrewr concept, going into full reduction.");
                    }
                    foundOneBadForkrewriting = true;
                    break outer;
                }
                case -1: {
                    System.out.println("TIMEOUT Gamma in optimization. Avoiding full reduction.");
                    return optiRetval;
                }
                case -2: {
                    System.out.println("TIMEOUT Omega in optimization. Avoiding full reduction.");
                    return optiRetval;
                }
                case 0: {
                    // Each fork rewriting gives us one intermediate program:
                    DatalogProgram progForP = tqCQrewr.getResultingProgram();
                    // We make it unique by replacing  Q(...)  by  P42(...)
                    progForP = replaceQbyPcounter(progForP);
                    if (CQContext.getINSTANCE().isVerbose()) {
                        System.out.println("Rewriting for p:\n" + progForP);
                    }
                    // Store it in the map for later use:
                    forkrewr2DlogMap.put(p, progForP);
                    break;
                }
                default: {
                    throw new CQException("[" + this.getClass().getSimpleName()
                            + "] Unexpected return value: " + optiRetval);
                }
            }
        }

        if (!foundOneBadForkrewriting) {
            if (CQContext.getINSTANCE().isVerbose()) {
                System.out.println("rCQ Optimization successful. Producing rewriting.");
            }

            // Construct the rewriting!
            DatalogProgram optiProgram = new DatalogProgram();

            for (RootedCQ p : allForkrewrOfQ0) {
                DatalogProgram progForP = forkrewr2DlogMap.get(p);
                // Put in all rules from p program:
                optiProgram.getSetOfRules().addAll(progForP.getSetOfRules());
                // Need one more rule  Q(...) <- P42(...)
                DatalogQueryAtom newQueryHead = new DatalogQueryAtom(avarsAsDatalogVariables);
                Set<DatalogAtom> newBodyAtoms = new HashSet<>();

                // Find the P42(..) rule:
                for (DatalogRule rule : progForP.getSetOfRules()) {
                    if (rule.getHeadAtom() instanceof DatalogQueryAtom) {
                        newBodyAtoms.add(rule.getHeadAtom());
                        break;
                    }
                }
                DatalogRule qFromP = new DatalogRule(newQueryHead, newBodyAtoms);
                optiProgram.getSetOfRules().add(qFromP);
            }

            this.resultingProgram = optiProgram;
            return 0;
        }

        //
        // Going into full Reduction:
        //
        if (CQContext.getINSTANCE().isVerbose()) {
            System.out.println("### rCQ, number of fork rewritings: " + this.getNumberOfFRs() + " = calls to bc+AQ");
            System.out.println("###      number of splittings:      " + this.getNumberOfSplittings() + " (only needed in case we do Omega!)");
        }

        if (CQContext.getINSTANCE().isVerbose()) {
            System.out.println("q0:  " + q0);
            System.out.println("Constructing Tp [and potentially Tpmin] for each fork rewriting p of q0...");
        }

        Set<Splitting> setOfAllSplittings = new HashSet<>();

        for (RootedCQ p : allForkrewrOfQ0) {

            long t1 = System.currentTimeMillis();

            if (CQContext.getINSTANCE().isVerbose()) {
                System.out.println("--- Fork rewriting p: " + p);
            }

            // Generate ontology Tp for the Gamma phase:
            Set<OWLAxiom> ontTpAxioms = RootedReasoner.getOntologyTpForP(this.getAxioms(), p);
            OWLOntology ontTp = OWLManager.createOWLOntologyManager().createOntology(ontTpAxioms);

            if (CQContext.getINSTANCE().isDEBUG()) {
                System.out.println("Have to check matches of all f.r. q into p.");
            }

            // Now for Tmin, the ontology for the Omega phase.
            // Tmin is built from Tp, and then extended by a CI for 
            // each splitting of each fork rewriting q of q0 wrt. p
            Set<Splitting> allSplittingsForAllFRs = new HashSet<>();
            for (RootedCQ q : allForkrewrOfQ0) {

                //System.out.println("------ Fork rewriting q: " + q);
                // Here we have to check whether q can be disregarded.
                // The rule is to disregard q if it eliminates MORE forks
                // than p!
                if (q.identifiesAvarsWhichAreSeparateIn(p)) {
                    //System.out.println("Disregard q! It identifies more avars than p!");
                    continue;
                }

                //System.out.println("Computing all Partitions of q.");
                Set<Partition> allPartitionsOfQ = q.getAllPartitions();

                for (Partition partitionOfQ : allPartitionsOfQ) {
                    if (CQContext.getINSTANCE().isDEBUG()) {
                        System.out.println("Current Partition: " + partitionOfQ);
                    }
                    // Generate all valid splittings from part:
                    Set<Splitting> splittingsOfQPartitionToP = partitionOfQ.generateSplittings(p);

                    if (CQContext.getINSTANCE().isDEBUG()) {
                        System.out.println("Splittings of Partition to p: " + splittingsOfQPartitionToP);
                    }

                    //System.out.println("Found following Splittings to p: " + splittingsOfQtoP);
                    allSplittingsForAllFRs.addAll(splittingsOfQPartitionToP);
                }
            }

            setOfAllSplittings.addAll(allSplittingsForAllFRs);

            // Can start Gamma with Tp and Omega with Tpmin:
            TpGammaTpminOmegaCaller caller = new TpGammaTpminOmegaCaller(ontTp, allSplittingsForAllFRs);
            if (CQContext.getINSTANCE().isVerbose()) {
                System.out.println("Start Gamma with ontology of size: " + ontTp.getAxioms().size());
            }
            int retval = caller.runSingleConceptname(CQContext.getINSTANCE().getClsN());
//            int retval = 0;

            // Retval is -1 (timeout), -2 (Omega cycle), or 0 (FOrewr)
            switch (retval) {
                case -1:
                    if (CQContext.getINSTANCE().isVerbose()) {
                        System.out.println("(rCQ full red) Gamma TIMEOUT with p: " + p);
                    }
                    return retval;
                case -2:
                    if (CQContext.getINSTANCE().isVerbose()) {
                        System.out.println("(rCQ full red) Omega TIMEOUT with p: " + p);
                    }
                    return retval;
                case 1:
                    if (CQContext.getINSTANCE().isVerbose()) {
                        System.out.println("(rCQ full red) Omega CYCLIC (FOno) with p: " + p);
                    }
                    return retval;
                default:
                    break;
            }
        }

        this.numberOfSplittings = setOfAllSplittings.size();

        if (CQContext.getINSTANCE().isVerbose()) {
            System.out.println("All N: FOyes! Query is FO-rewritable!");
        }
        return 0;
    }

    private DatalogProgram replaceQbyPcounter(DatalogProgram progForP) {
        DatalogProgram prog = new DatalogProgram();
        String pPredicate = "P" + CQContext.getINSTANCE().getCounter();
        for (DatalogRule rule : progForP.getSetOfRules()) {
            DatalogAtom head = rule.getHeadAtom();
            if (head instanceof DatalogQueryAtom) {
                DatalogQueryAtom queryHead = (DatalogQueryAtom) head;
                Set<DatalogVariable> setOfVars = queryHead.getSetOfVars();
                DatalogQueryAtom newHead = new DatalogQueryAtom(setOfVars, pPredicate);
                Set<DatalogAtom> bodyAtoms = rule.getBody();

                DatalogRule newRule = new DatalogRule(newHead, bodyAtoms);
                prog.getSetOfRules().add(newRule);
            } else {
                prog.getSetOfRules().add(rule);
            }
        }
        return prog;
    }

}
