/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.cq;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import de.unibremen.informatik.tdki.grind2.cq.components.RoleAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.ClassAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.Atom;
import de.unibremen.informatik.tdki.grind2.utilities.ConceptTreeNode;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import org.jgrapht.DirectedGraph;
import org.jgrapht.UndirectedGraph;
import org.jgrapht.alg.ConnectivityInspector;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.graph.DefaultEdge;
import org.jgrapht.graph.DirectedPseudograph;
import org.jgrapht.graph.Pseudograph;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLObjectProperty;

/**
 *
 * @author hansen
 */
public class CQ {

    final private static double ATOMPROB = 0.1;
    final private static double AVARPROB = 0.4;

    final public static int GENERALCQ = 0;
    final public static int ROOTEDCQ = 1;
    final public static int TQCQ = 2;
    final public static int AVARCQ = 3;
    final public static int CONCQ = 4;
    final public static int AQ = 5;

//    private CQ coreOnlyRoles;
    public static double getATOMPROB() {
        return ATOMPROB;
    }

    public static double getAVARPROB() {
        return AVARPROB;
    }

    final private Set<Atom> setOfAtoms;
    Set<Variable> avars;
    Set<Variable> vars;

    public CQ(Set<Variable> avars, Set<Atom> setOfAtoms) {
        this.setOfAtoms = new HashSet(setOfAtoms);
        this.avars = new HashSet(avars);

        this.vars = new HashSet<>();
        for (Atom a : this.setOfAtoms) {
            this.vars.addAll(a.getVars());
        }
    }

    public Set<Variable> getAvars() {
        return avars;
    }

    public Set<Atom> getSetOfAtoms() {
        return setOfAtoms;
    }

    public final Set<Variable> getVars() {
        return this.vars;
    }

    public int getType() throws CQException {
        if (!this.isConnected()) {
            return GENERALCQ;
        }
        if (this.getAvars().isEmpty()) {
            return GENERALCQ;
        }
        if (this.isAQ()) {
            return AQ;
        }
        if (this.isConCQ()) {
            return CONCQ;
        }
        if (this.isAvarCQ()) {
            return AVARCQ;
        }
        if (this.isTqCQ()) {
            return TQCQ;
        }
        if (this.isRootedCQ()) {
            return ROOTEDCQ;
        }
        return GENERALCQ;
    }

    public boolean isAQ() {
        if (avars.size() != 1) {
            return false;
        }
        if (setOfAtoms.size() != 1) {
            return false;
        }
        Atom singleAtom = setOfAtoms.iterator().next();
        if (singleAtom instanceof RoleAtom) {
            return false;
        }
        return true;
    }

    public AQ asAQ() throws CQException {
        return new AQ(this.getAvars(), this.getSetOfAtoms());
    }

    public boolean isConCQ() {
        if (this.getAvars().size() != 1) {
            return false;
        }
        Variable root = this.getAvars().iterator().next();
        return this.isTreeCQWithRoot(root);
    }

    public ConCQ asConCQ() throws CQException {
        Variable root = this.getAvars().iterator().next();
        return new ConCQ(this.getSetOfAtoms(), root);
    }

    /**
     * Checks whether this CQ is actually a TreeCQ rooted in root.
     *
     * @param root
     * @return
     */
    public boolean isTreeCQWithRoot(Variable root) {
        // root has to have indegree of zero:
        if (this.indegree(root) != 0) {
            return false;
        }
        // Others must have indegree of one:
        Set<Variable> varsWithoutRoot = new HashSet<>(this.getVars());
        varsWithoutRoot.remove(root);
        for (Variable x : varsWithoutRoot) {
            if (this.indegree(x) != 1) {
                return false;
            }
        }
        // Trees have to be connected:
        if (!this.isConnected()) {
            return false;
        }
        // Check that we don't have multiple edges r(x,y), s(x,y):
        Set<List<Variable>> setOfXandY = new HashSet<>();
        for (Atom at : this.setOfAtoms) {
            if (at instanceof RoleAtom) {
                List<Variable> xAndY = new ArrayList<>();
                xAndY.add(((RoleAtom) at).getXVar());
                xAndY.add(((RoleAtom) at).getYVar());
                if (setOfXandY.contains(xAndY)) {
                    return false;
                } else {
                    setOfXandY.add(xAndY);
                }
            }
        }
        // And they have to be acyclic:
        if (!this.isAcyclic()) {
            return false;
        }
        return true;
    } // isTreeWithRoot

    public TreeCQ asTreeCQWithRoot(Variable root) throws CQException {
        return new TreeCQ(this.getSetOfAtoms(), root);
    }

    /**
     * Checks whether this CQ is actually an AvarCQ.
     *
     * @return
     */
    public boolean isAvarCQ() {
        // Check we have at least one avar:
        if (avars.isEmpty()) {
            return false;
        }

        // Check all used vars are avars:
        Set<Variable> setOfAllVars = new HashSet<>();
        for (Atom a : setOfAtoms) {
            setOfAllVars.addAll(a.getVars());
        }
        if (!setOfAllVars.equals(avars)) {
            return false;
        }
        return true;
    }

    public AvarCQ asAvarCQ() throws CQException {
        return new AvarCQ(this.getAvars(), this.getSetOfAtoms());
    }

    public boolean isTqCQ() throws CQException {
        CQ core = this.getCore();
        //System.out.println(core);
        if (!core.getVars().equals(this.getAvars())) {
            return false;
        }
        return true;
    }

    public TqCQ asTqCQ() throws CQException {
        return new TqCQ(this.getAvars(), this.getSetOfAtoms());
    }

    /**
     * Checks whether this CQ is actually a rooted CQ.
     *
     * @return
     */
    public boolean isRootedCQ() {
        if (this.getAvars().isEmpty()) {
            return false;
        }
        if (!this.isConnected()) {
            return false;
        }
        return true;
    }

    public RootedCQ asRootedCQ() throws CQException {
        return new RootedCQ(this.getAvars(), this.getSetOfAtoms());
    }

    public final boolean isConnected() {
        UndirectedGraph<Variable, DefaultEdge> graph = new Pseudograph<>(DefaultEdge.class);
        for (Atom at : this.setOfAtoms) {
            if (at instanceof ClassAtom) {
                graph.addVertex(((ClassAtom) at).getVar());
            } else if (at instanceof RoleAtom) {
                RoleAtom rat = (RoleAtom) at;
                graph.addVertex(rat.getXVar());
                graph.addVertex(rat.getYVar());
                graph.addEdge(rat.getXVar(), rat.getYVar());
            }
        }
        if (graph.vertexSet().isEmpty()) {
            return true;
        }
        ConnectivityInspector ci = new ConnectivityInspector(graph);
        return ci.isGraphConnected();
    }

    public final boolean isAcyclic() {
        DirectedGraph<Variable, DefaultEdge> graph = new DirectedPseudograph<>(DefaultEdge.class);
        for (Atom at : this.setOfAtoms) {
            //if (at instanceof ClassAtom) {
            //    graph.addVertex(((ClassAtom) at).getVar());
            //} else 
            if (at instanceof RoleAtom) {
                graph.addVertex(((RoleAtom) at).getXVar());
                graph.addVertex(((RoleAtom) at).getYVar());
                graph.addEdge(((RoleAtom) at).getXVar(), ((RoleAtom) at).getYVar());
            }
        }
        CycleDetector ci = new CycleDetector(graph);
        return !ci.detectCycles();
    }

    public final int outdegree(Variable x) {
        int result = 0;
        for (Atom at : this.setOfAtoms) {
            if (at instanceof RoleAtom) {
                if (((RoleAtom) at).getXVar().equals(x)) {
                    result++;
                }
            }
        }
        return result;
    }

    public final int indegree(Variable x) {
        int result = 0;
        for (Atom at : this.setOfAtoms) {
            if (at instanceof RoleAtom) {
                if (((RoleAtom) at).getYVar().equals(x)) {
                    result++;
                }
            }
        }
        return result;
    }

    /**
     * Proposed algorithm to find core and trees. - Discard class atoms A(x). -
     * DO: - find xVars, Multiset - find yVars, Multiset - do Ys \setminus Xs -
     * find singletons - remove atoms with singleton as Y, but no
     * avar/avar-containing var - UNTIL nothing changed.
     *
     * @return
     * @throws de.unibremen.informatik.tdki.grind2.cq.CQException
     */
    public CQ getCore() throws CQException {

//        if (this.coreOnlyRoles != null) {
//            return this.coreOnlyRoles;
//        }
        Set<Atom> newAtomSet = new HashSet<>(this.setOfAtoms);

        // Remove class atoms:
        Set<Atom> removeTheseAtoms = new HashSet<>();
        for (Atom atom : newAtomSet) {
            if ((atom instanceof ClassAtom)) {
                removeTheseAtoms.add(atom);
            }
        }
        newAtomSet.removeAll(removeTheseAtoms); // cannot do concurrently!

        // Delete role atoms that form trees, bottom-up.
        boolean atomRemoved;
        do {
            atomRemoved = false;

            Multiset<Variable> xVars = HashMultiset.create();
            Multiset<Variable> yVars = HashMultiset.create();
            Map<Variable, RoleAtom> y2ratMap = new HashMap<>();
            Set<Variable> nonSingletonYVars = new HashSet<>();
            for (Atom at : newAtomSet) {
                if ((at instanceof RoleAtom)) {
                    RoleAtom rat = (RoleAtom) at;
                    Variable x = rat.getXVar();
                    Variable y = rat.getYVar();
                    xVars.add(x);
                    yVars.add(y);
                    if (y2ratMap.containsKey(y)) {
                        nonSingletonYVars.add(y);
                    } else {
                        y2ratMap.put(y, rat);
                    }
                }
            }

            // remove all xVars, avars, and non-singletons from Ys:
            yVars.removeAll(xVars);

            for (Variable x : this.getVars()) {
                if (x.containsSomeAvarOf(this.avars)) {
                    yVars.remove(x);
                }
            }

            yVars.removeAll(nonSingletonYVars);

            // delete atoms that contain okayToDeleteYVars as y:
            for (Variable y : yVars) {
                if (newAtomSet.remove(y2ratMap.get(y))) {
                    atomRemoved = true;
                }
            }

        } while (atomRemoved);

        // Find out which vars are core vars:
        Set<Variable> coreVars = new HashSet<>();
        for (Atom at : newAtomSet) {
            coreVars.addAll(at.getVars());
        }

        // Might have lost some avar/avar-containing vars:
        for (Variable x : this.getVars()) {
            if (x.containsSomeAvarOf(this.getAvars())) {
                coreVars.add(x);
            }
        }

        // Now fill core again with class atoms:
        for (Atom at : this.setOfAtoms) {
            if ((at instanceof ClassAtom)) {
                ClassAtom cat = (ClassAtom) at;
                if (coreVars.contains(cat.getVar())) {
                    newAtomSet.add(cat);
                }
            }
        }

        // Can return a query with same avars, reduced atom set:
        CQ core = new CQ(this.avars, newAtomSet);
        return core;
    }

    /**
     * Compute "finest" Partition, where only the core is R.
     *
     * @return
     * @throws CQException
     */
    public Partition getFinestPartition() throws CQException {
        Set<Variable> emptyset = new HashSet<>();
        return this.getPartition(emptyset);
    }

    /**
     * Return all Partitions. They are generated by taking all possible
     * combinations of non-core variables additionally into R (powerset
     * command). "Possible" = valid R set, i.e. not allowed is: non-R-var -->
     * R-var edge.
     *
     * @return
     * @throws CQException
     */
    public Set<Partition> getAllPartitions() throws CQException {

        Set<Partition> result = new HashSet<>();
        Partition finestPartition = this.getFinestPartition();

        Set<Variable> allVarsInTrees = new HashSet<>();
        for (TreeCQ tree : finestPartition.getTrees()) {
            allVarsInTrees.addAll(tree.getVars());
        }

        for (Set<Variable> oneTreeVarSet : Sets.powerSet(allVarsInTrees)) {
            Partition partForThisSet = getPartition(oneTreeVarSet);
            if (partForThisSet == null) {
                // Not a valid R set!
            } else {
                result.add(partForThisSet);
            }
        }
        return result;
    }

    /**
     * Return Partition into R,S1...Sn wrt a set R of variables. The partition
     * can be extended to a splitting.
     *
     * @param additionalVariablesInR
     * @return
     * @throws CQException
     */
    private Partition getPartition(Set<Variable> additionalVariablesInR) throws CQException {

        // Start with core:
        CQ core = this.getCore();
        Set<Atom> rAtomSet = new HashSet<>(core.getSetOfAtoms());
        Set<Variable> coreVars = core.getVars();

        //Set<Variable> coreVars = Sets.union(rAtomsCQ.getVars(), this.getAvars());
        Set<Variable> rVars = Sets.union(additionalVariablesInR, coreVars);

        // Datastructures for r(x,y) with x in R, y notin R:
        List<Variable> treeRootParents = new ArrayList<>();
        List<Variable> treeRoots = new ArrayList<>();
        List<OWLObjectProperty> treeRootRoles = new ArrayList<>();
        List<TreeCQ> trees = new ArrayList<>();

        // Add classAtoms for rVars, and register TreeRoot situations.
        // At the same time, we have to ensure that every R variable appears 
        // in at least one atom. 
        Set<Variable> rVarAppeared = new HashSet<>();
        for (Atom at : this.getSetOfAtoms()) {
            if (at instanceof ClassAtom) {
                ClassAtom cat = (ClassAtom) at;
                Variable x = cat.getVar();
                if (rVars.contains(x)) {
                    rAtomSet.add(at);
                    rVarAppeared.add(x);
                }

            } else if (at instanceof RoleAtom) {
                RoleAtom rat = (RoleAtom) at;
                Variable x = rat.getXVar();
                Variable y = rat.getYVar();

                if (!rVars.contains(x)) {
                    if (rVars.contains(y)) {
                        return null;
                    }
                }

                if (rVars.contains(x)) {
                    rVarAppeared.add(x);
                    if (rVars.contains(y)) {

                        // We might not have this roleAtom in our R set:
                        rAtomSet.add(at);
                        rVarAppeared.add(y);

                    } else if (!rVars.contains(y)) {

                        // We have TreeRoot situation:
                        treeRootParents.add(x);
                        treeRoots.add(y);
                        treeRootRoles.add(rat.getRole());
                        trees.add(this.getTreeCQInAcyclicPartAt(y));
                    }
                }
            }
        }

        // Check if all R vars appeared; if not, add Top(x).
        for (Variable x : rVars) {
            if (!rVarAppeared.contains(x)) {
                rAtomSet.add(
                        new ClassAtom(OWLManager.getOWLDataFactory().getOWLThing(), x));
            }
        }

        CQ newRAtomCQ = new CQ(this.getAvars(), rAtomSet);
        return new Partition(newRAtomCQ, treeRootParents, treeRoots, treeRootRoles, trees);
    }

    public TreeCQ getTreeCQInAcyclicPartAt(Variable x) throws CQException {
        // Visit all variables:
        Map<Variable, ConceptTreeNode> vars2TreeNodeMap = new HashMap<>();
        fillTreeNodes(vars2TreeNodeMap);

        // Check that we have something in the map at x:
        if (vars2TreeNodeMap.containsKey(x)) {
            // Choose our root as starting TreeNode:
            ConceptTreeNode rootTreeNode = vars2TreeNodeMap.get(x);
            Set<Atom> gatheredAtomSet = gatherSubtreeAtoms(rootTreeNode);
            if (gatheredAtomSet.isEmpty()) {
                gatheredAtomSet.add(new ClassAtom(OWLManager.getOWLDataFactory().getOWLThing(), x));
            }
            return new TreeCQ(gatheredAtomSet, x);
        } else {
            return new TreeCQ(new HashSet<Atom>(), x);
        }
    }

    public void fillTreeNodes(Map<Variable, ConceptTreeNode> vars2TreeNodeMap) throws CQException {

        // Process all atoms:
        for (Atom at : this.getSetOfAtoms()) {
            if (at instanceof ClassAtom) {
                // It's a class atom:
                ClassAtom clsAtom = (ClassAtom) at;
                Variable x = clsAtom.getVar();
                if (vars2TreeNodeMap.containsKey(x)) {
                    vars2TreeNodeMap.get(x).addClass(clsAtom.getCls());
                } else {
                    ConceptTreeNode tn = new ConceptTreeNode(x);
                    tn.addClass(clsAtom.getCls());
                    vars2TreeNodeMap.put(x, tn);
                }

            } else if (at instanceof RoleAtom) {
                // It's a role atom:
                RoleAtom rat = (RoleAtom) at;
                Variable x = rat.getXVar();
                Variable y = rat.getYVar();
                ConceptTreeNode xNode;
                ConceptTreeNode yNode;
                if (vars2TreeNodeMap.containsKey(x)) {
                    xNode = vars2TreeNodeMap.get(x);
                } else {
                    xNode = new ConceptTreeNode(x);
                    vars2TreeNodeMap.put(x, xNode);
                }
                if (vars2TreeNodeMap.containsKey(y)) {
                    yNode = vars2TreeNodeMap.get(y);
                } else {
                    yNode = new ConceptTreeNode(y);
                    vars2TreeNodeMap.put(y, yNode);
                }
                xNode.addChild(yNode, rat.getRole());

            } else {
                // Other kinds of atoms:
                throw new CQException("[" + this.getClass().getSimpleName()
                        + "] Tree building found non-role and non-class atom: " + at);
            }
        }
    }

    private Set<Atom> gatherSubtreeAtoms(ConceptTreeNode xNode) throws CQException {
        Set<Atom> resultSetOfAtoms = new HashSet<>();

        Variable x = xNode.getX();
        for (OWLClass cls : xNode.getSetOfClasses()) {
            resultSetOfAtoms.add(new ClassAtom(cls, x));
        }
        if (!xNode.isLeaf()) {
            for (ConceptTreeNode child : xNode.getChildren()) {
                resultSetOfAtoms.addAll(gatherSubtreeAtoms(child));
                resultSetOfAtoms.add(new RoleAtom(child.getReachingRole(), x, child.getX()));
            }
        }
        return resultSetOfAtoms;
    }

    @Override
    public String toString() {
        List sortedList = new ArrayList(avars);
        Collections.sort(sortedList);
        return "q" + sortedList + " = " + this.setOfAtoms;
    }

    public String toDatalogString() {
        List sortedList = new ArrayList(avars);
        Collections.sort(sortedList);
        String head = sortedList.toString().replaceFirst("\\[", "Q(").replaceAll("\\]", "").replaceAll("x\\[", "\\?") + ")";
        String body = this.setOfAtoms.toString().replaceAll("x\\[", "?").replaceAll("\\[", "").replaceAll("\\]", "").replaceAll("<", "").replaceAll(">", "");
        return head + "  <-  " + body;
    }

    @Override
    public int hashCode() {
        int hash = 5;
        hash = 53 * hash + Objects.hashCode(this.setOfAtoms);
        hash = 53 * hash + Objects.hashCode(this.avars);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (this.getClass().isAssignableFrom(obj.getClass())) {

            final CQ other = (CQ) obj;
            if (!Objects.equals(this.setOfAtoms, other.setOfAtoms)) {
                return false;
            }
            if (!Objects.equals(this.avars, other.avars)) {
                return false;
            }
            return true;
        }
        return false;
    }

}
