/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.cq;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLObjectProperty;

/**
 *
 * @author hansen
 */
public class Partition {

    private final CQ rPart;
    private final List<Variable> treeRootParents;
    private final List<Variable> treeRoots;
    private final List<OWLObjectProperty> treeRootRoles;
    private final List<TreeCQ> trees;
    private final Set<Variable> identifiedAvarsInR;
    private final Set<Integer> allAvarIndices;

    public Partition(CQ rPart, List<Variable> treeRootParents, List<Variable> treeRoots, List<OWLObjectProperty> treeRootRoles, List<TreeCQ> trees) {
        this.rPart = rPart;
        this.treeRootParents = treeRootParents;
        this.treeRoots = treeRoots;
        this.treeRootRoles = treeRootRoles;
        this.trees = trees;
        this.allAvarIndices = new HashSet<>();
        for (Variable x : this.rPart.getAvars()) {
            allAvarIndices.addAll(x.getIndices());
        }
        identifiedAvarsInR = new HashSet<>();
        fillIdAvarsInR();
    }

    /**
     * Avars of R are x0, x1, but identified avars are e.g x01.
     */
    private void fillIdAvarsInR() {
        for (Variable x : this.getrPart().getVars()) {
            if (!Sets.intersection(x.getIndices(), allAvarIndices).isEmpty()) {
                this.identifiedAvarsInR.add(x);
            }
        }

    }

    public CQ getCore() {
        return rPart;
    }

    public Set<Integer> getAllAvarIndices() {
        return allAvarIndices;
    }

    public List<Variable> getTreeRootParents() {
        return treeRootParents;
    }

    public List<Variable> getTreeRoots() {
        return treeRoots;
    }

    public List<OWLObjectProperty> getTreeRootRoles() {
        return treeRootRoles;
    }

    public List<TreeCQ> getTrees() {
        return trees;
    }

    public CQ getrPart() {
        return rPart;
    }

    public Map<Variable, Variable> computeR2pAvarMap(RootedCQ p) throws CQException {
        Map<Variable, Variable> r2pAvarMap = new HashMap<>();

        // Set of all avar indices in p:
        Set<Integer> pAvarIndices = new HashSet<>();
        for (Variable x : p.getAvars()) {
            pAvarIndices.addAll(x.getIndices());
        }

        outer:
        for (Variable rVar : this.identifiedAvarsInR) {
            for (Variable pVar : p.getVars()) {
                Set<Integer> pVarAvarIndices = pVar.getAvarIndices(pAvarIndices);
                Set<Integer> rVarAvarIndices = rVar.getAvarIndices(pAvarIndices);
                if (pVarAvarIndices.containsAll(rVarAvarIndices)) {
                    r2pAvarMap.put(rVar, pVar);
                    continue outer;
                }
            }
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Could not find matching p variable for " + rVar + "!");
        }
        return r2pAvarMap;
    }

    /**
     * From this Partition, we can generate Splittings wrt p. The Splittings are
     * all possibilities that map R to ABox Ap.
     *
     * @param p
     * @return
     * @throws de.unibremen.informatik.tdki.grind2.cq.CQException
     */
    public Set<Splitting> generateSplittings(RootedCQ p) throws CQException {

        Map<Variable, Variable> r2pAvarMap = this.computeR2pAvarMap(p);

        // The targets of the Map: all Vars in p:
        List<Variable> listOfTargets = Lists.newArrayList(p.getVars());

        // Each source R var is a "position"
        List<Variable> listOfPositions = Lists.newArrayList(this.getrPart().getVars());
        int numberOfPositions = listOfPositions.size();

        // Master map, stores fixed maps of R avars to their counterpart in p:
        Variable[][] positions2Targets = new Variable[2][numberOfPositions];

        // Check which positions are free and which are fixed:
        List<Integer> listOfFreePositions = new ArrayList<>();
        Map<Integer, Integer> position2FreePositionMap = new HashMap<>();
        int numberOfFreePositions = 0;
        for (int i = 0; i < numberOfPositions; i++) {
            Variable x = listOfPositions.get(i);
            positions2Targets[0][i] = x;
            if (this.identifiedAvarsInR.contains(x)) {
                // This position is fixed. Fill target:
                if (!r2pAvarMap.containsKey(x)) {
                    throw new CQException("[" + this.getClass().getSimpleName()
                            + "] Could not find matching p variable for " + x + "!");

                }
                positions2Targets[1][i] = r2pAvarMap.get(x);
            } else {
                // This position is free in the Map:
                listOfFreePositions.add(i);
                position2FreePositionMap.put(i, numberOfFreePositions);
                numberOfFreePositions++;
            }
        }
//        System.out.println("Mapping:");
//        for (int i = 0; i < numberOfPositions; i++) {
//            System.out.println("[0][" + i + "] = " + positions2Targets[0][i]
//                    + " [1][" + i + "] = " + positions2Targets[1][i]);
//        }

        // Compute the number of Splittings we get here; it's exponential in
        // the number of free Variables in q:
        int numberOfTargets = listOfTargets.size();
        int numberOfPVarsTOTHEnumberOfFreePositions
                = (int) Math.pow(numberOfTargets, numberOfFreePositions);

        Set<Splitting> resultSplittings = new HashSet<>();
        for (int counterBaseN = 0;
                counterBaseN < numberOfPVarsTOTHEnumberOfFreePositions;
                counterBaseN++) {

            // We need only nu to create a Splitting:
            Map<Variable, Variable> nu = new HashMap<>();

            // Magic part, generate mapping for
            //  - fixed positions (already done),
            //  - free positions by using counterBaseN:
            for (int j = 0; j < numberOfPositions; j++) {

                // Check if we are at a free position:
                if (listOfFreePositions.contains(j)) {

                    int freePosition = position2FreePositionMap.get(j);

                    // Map is not in our array:
                    Variable qVar = positions2Targets[0][j];

                    // Generate power numbers to extract digit for position:
                    int powernumberForFreeposition = (int) Math.pow(numberOfTargets, freePosition);
                    int counterDigitForFreepos = (counterBaseN / powernumberForFreeposition) % numberOfTargets;
                    Variable pVar = listOfTargets.get(counterDigitForFreepos);

                    nu.put(qVar, pVar);
                } else {
                    // Fixed position! Map is already in our array:
                    Variable qVar = positions2Targets[0][j];
                    Variable pVar = positions2Targets[1][j];

                    nu.put(qVar, pVar);
                }
            }
            Splitting newSplit = new Splitting(this, nu);
            if (newSplit.atomsExistingIn(p)) {
                resultSplittings.add(newSplit);
            } else {
                // Didn't find all roles for newSplit.
            }
        }
        return resultSplittings;
    }

    @Override
    public String toString() {
        return "Partition{rPart=" + rPart + ", treeRootParents="
                + treeRootParents + ", treeRoots=" + treeRoots
                + ", treeRootRoles=" + treeRootRoles + ", trees=" + trees + '}';
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 29 * hash + Objects.hashCode(this.rPart);
        hash = 29 * hash + Objects.hashCode(this.treeRootParents);
        hash = 29 * hash + Objects.hashCode(this.treeRoots);
        hash = 29 * hash + Objects.hashCode(this.treeRootRoles);
        hash = 29 * hash + Objects.hashCode(this.trees);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Partition other = (Partition) obj;
        if (!Objects.equals(this.rPart, other.rPart)) {
            return false;
        }
        if (!Objects.equals(this.treeRootParents, other.treeRootParents)) {
            return false;
        }
        if (!Objects.equals(this.treeRoots, other.treeRoots)) {
            return false;
        }
        if (!Objects.equals(this.treeRootRoles, other.treeRootRoles)) {
            return false;
        }
        if (!Objects.equals(this.trees, other.trees)) {
            return false;
        }
        return true;
    }

}
