/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.cq;

import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import de.unibremen.informatik.tdki.grind2.utilities.Signature;
import de.unibremen.informatik.tdki.grind2.cq.components.RoleAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.ClassAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.Atom;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Sets;
import de.unibremen.informatik.tdki.grind2.utilities.SimpleOntology;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLObjectProperty;

/**
 *
 * @author hansen
 */
public class RootedCQ extends CQ {

    public RootedCQ(Set<Variable> avars, Set<Atom> setOfAtoms) throws CQException {
        super(avars, setOfAtoms);

        // Need to have some avars:
        if (avars.isEmpty()) {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried to construct RootedCQ without avars!");
        }

        // Need to be connected:
        if (!this.isConnected()) {
            throw new CQException("[" + this.getClass().getSimpleName()
                    + "] Tried to construct non-connected RootedCQ!");
        }
    }

    public RootedCQ getMinimalCQ(SimpleOntology t) {

        return null;

    }

    public Set<RootedCQ> getForkRewritings() throws CQException {

        Set<RootedCQ> readyForkRewritings = new HashSet<>();
        Set<RootedCQ> toDoQueries = new HashSet<>();
        toDoQueries.add(this);

        while (!toDoQueries.isEmpty()) {
            // Take one from set:
            RootedCQ queryToProcess = toDoQueries.iterator().next();

            toDoQueries.remove(queryToProcess);
            readyForkRewritings.add(queryToProcess);
            CQ core = queryToProcess.getCore();
            if (!core.isRootedCQ()) {
                throw new CQException("[" + RootedCQ.class.getSimpleName()
                        + "] getForkRewritings, found core that is not a RootedCQ!");
            }
            Set<Fork> setOfForks = core.asRootedCQ().findForks();

            Set<RootedCQ> setOfForkrewritings = queryToProcess.allForkrewrFromPowerset(setOfForks);

//            System.out.println("Found fork rewr: " + setOfForkrewritings.toString().replace("q", "\nq"));
            setOfForkrewritings.removeAll(readyForkRewritings);
            toDoQueries.addAll(setOfForkrewritings);
        }

        return readyForkRewritings;
    }

    private Set<RootedCQ> allForkrewrFromPowerset(Set<Fork> setOfForks) throws CQException {

        Set<RootedCQ> resultSet = new HashSet<>();

        // Every set of forks in the power set of the fork set represents a valid new CQ.
        // Iterate over all of them:
        for (Set<Fork> forkSet : Sets.powerSet(setOfForks)) {

            // Empty fork set: ignore.
            if (forkSet.isEmpty()) {
                continue;
            }

            // New Atom set for fork rewriting:
            Set<Atom> copyOfSetOfAtoms = new HashSet<>(this.getSetOfAtoms());
            Set<Atom> newSetOfAtoms = new HashSet<>(this.getSetOfAtoms());

            // Need to check whether some forks have overlapping variables:
            Map<Fork, Variable> fork2newSourceVarMap = new HashMap<>();
            fillfork2newSourceVarMap(fork2newSourceVarMap, forkSet);

            // Execute every fork:
            for (Fork fork : forkSet) {
                Variable newSourceVar = fork2newSourceVarMap.get(fork);

                // Change all atoms to use the identified variable:
                for (Atom at : copyOfSetOfAtoms) {
                    // There might be overlap between current atom and sourceVars:
                    // If yes, build a new Atom!
                    newSetOfAtoms.remove(at);
                    Atom newAtom = at.replaceIdentifiedVars(newSourceVar);
                    newSetOfAtoms.add(newAtom);
                }
            }
            resultSet.add(new RootedCQ(this.getAvars(), newSetOfAtoms));
        }
        return resultSet;
    }

    /**
     * This is to check that two overlapping forks get assigned the same
     * Variable, a union of both sourceVars.
     *
     * @param fork2newSourceVarMap
     * @param forkSet
     */
    private void fillfork2newSourceVarMap(Map<Fork, Variable> fork2newSourceVarMap, Set<Fork> forkSet) throws CQException {

        for (Fork fork : forkSet) {

            // If this fork already appeared, continue:
            if (fork2newSourceVarMap.containsKey(fork)) {
                continue;
            }

            for (Fork fork2 : forkSet) {

                // If this fork2 already appeared, continue:
                if (fork2newSourceVarMap.containsKey(fork2)) {
                    continue;
                }

                // Check if fork and fork2 share variables:
                if (!Collections.disjoint(fork.getSourceVars(), fork2.getSourceVars())) {
                    // Overlapping forks! Generate union Variable:
                    Set<Variable> unionSet = new HashSet<>(fork.getSourceVars());
                    unionSet.addAll(fork2.getSourceVars());
                    Variable unionVar = new Variable(unionSet);
                    fork2newSourceVarMap.put(fork, unionVar);
                    fork2newSourceVarMap.put(fork2, unionVar);
                } else {
                    fork2newSourceVarMap.put(fork, new Variable(fork.getSourceVars()));
                    fork2newSourceVarMap.put(fork2, new Variable(fork2.getSourceVars()));
                }

            } // forks x forks
        } //     forks x forks
    }

    /**
     * Finds all relevant forks in a core of a RootedCQ. Algorithm: - find sinks
     * (leaves); - foreach sink: check incoming roles; - if >=2 and all same:
     * fork found!
     *
     * @param q
     * @return
     */
    private Set<Fork> findForks() throws CQException {
        Multiset<Variable> xVars = HashMultiset.create();
        Multiset<Variable> yVars = HashMultiset.create();
        Map<Variable, Set<OWLObjectProperty>> y2RolesMap = new HashMap<>();
        Map<Variable, Set<RoleAtom>> y2RoleAtomsMap = new HashMap<>();

        // Fill xVars, yVars, and Maps:
        fillXYMaps(xVars, yVars, y2RoleAtomsMap, y2RolesMap);

        // Sinks are y-only:
        Multiset<Variable> sinkVars = HashMultiset.create();
        sinkVars.addAll(yVars);
        sinkVars.removeAll(xVars);

        // Only sinks with unique incoming roles are right sinks;
        // also, there should be no leafs (count == 1):
        checkUniqueIncomingRole(sinkVars, y2RolesMap);

        // Avars cannot be sinks of a fork:
        Set<Variable> extendedAvarSet = new HashSet<>();
        for (Variable sink : sinkVars) {
            if (sink.containsSomeAvarOf(avars)) {
                extendedAvarSet.add(sink);
            }
        }
        sinkVars.removeAll(extendedAvarSet);

        Set<Fork> setOfForks = new HashSet<>();
        for (Variable sink : sinkVars) {
            Set<RoleAtom> forkAtoms;

            // Set should contain only one Role:
            OWLObjectProperty role = y2RolesMap.get(sink).iterator().next();
            forkAtoms = y2RoleAtomsMap.get(sink);

            // Get source Vars:
            Set<Variable> sourceVars = new HashSet<>();
            for (RoleAtom rat : forkAtoms) {
                sourceVars.add(rat.getXVar());
            }
            setOfForks.add(new Fork(//forkAtoms, 
                    role, sourceVars, sink));

        }
        return setOfForks;
    }

    private void checkUniqueIncomingRole(Multiset<Variable> sinkVars,
            Map<Variable, Set<OWLObjectProperty>> y2RolesMap) throws CQException {
        Set<Variable> removeSinks = new HashSet<>();
        for (Variable sink : sinkVars) {
            if (sinkVars.count(sink) == 1) {
                removeSinks.add(sink);
            }
            if (y2RolesMap.get(sink).size() > 1) {
                removeSinks.add(sink);
            }
        }
        sinkVars.removeAll(removeSinks);
    }

    private void fillXYMaps(Multiset<Variable> xVars, Multiset<Variable> yVars, Map<Variable, Set<RoleAtom>> y2RoleAtomsMap, Map<Variable, Set<OWLObjectProperty>> y2RolesMap) {
        for (Atom at : this.getSetOfAtoms()) {

            if (at instanceof RoleAtom) {

                // Fill XY Maps:
                RoleAtom rat = (RoleAtom) at;
                Variable x = rat.getXVar();
                Variable y = rat.getYVar();
                OWLObjectProperty role = rat.getRole();

                xVars.add(x);
                yVars.add(y);

                // Update y2RoleAtomsMap:
                if (y2RoleAtomsMap.containsKey(y)) {
                    y2RoleAtomsMap.get(y).add(rat);
                } else {
                    Set<RoleAtom> newRoleAtomSet = new HashSet<>();
                    newRoleAtomSet.add(rat);
                    y2RoleAtomsMap.put(y, newRoleAtomSet);
                }

                // Update y2RolesMap:
                if (y2RolesMap.containsKey(y)) {
                    y2RolesMap.get(rat.getYVar()).add(role);
                } else {
                    Set<OWLObjectProperty> newRoleSet = new HashSet<>();
                    newRoleSet.add(role);
                    y2RolesMap.put(y, newRoleSet);
                }
            }
        }
    } // fillXYMaps

    /**
     * Check whether during the creation of (this forkrewr q) some avars were
     * identified that are separate in the given forkrewr p. Stores fitting avar
     * pairs in q2pAvarMap.
     *
     * @param p
     * @param q2pAvarMap
     * @return
     */
    public boolean identifiesAvarsWhichAreSeparateIn(RootedCQ p) throws CQException {
        // Algorithm: Check that for every only-avar-IdxSet sQ in q, there is an 
        // only-avar-IdxSet sP in p such that sQ subseteq sP

        // Compute set of Indices of Non-Avars:
        Set<Integer> allIndices = new HashSet<>();
        Set<Integer> allAvarIndices = new HashSet<>();
        for (Variable x : this.getVars()) {
            allIndices.addAll(x.getIndices());
        }
        for (Variable x : this.getAvars()) {
            allAvarIndices.addAll(x.getIndices());
        }
        //System.out.println("allIndices: " + allIndices);
        //System.out.println("allAvarIndices: " + allAvarIndices);

        // Get all sP only-avar-IdxSets:
        Set<Set<Integer>> setOfAllPOnlyAvarIdxSets = new HashSet<>();
        Map<Set<Integer>, Variable> pIdxSet2VariableMap = new HashMap<>();
        for (Variable x : p.getVars()) {
            Set<Integer> sP = Sets.newHashSet(x.getIndices());
            sP.retainAll(allAvarIndices); // Only avars!
            setOfAllPOnlyAvarIdxSets.add(sP);
            // Add map entry only for avars:
            if (sP.size() > 0) {
                pIdxSet2VariableMap.put(sP, x);
            }
        }
        //System.out.println("All sP: " + setOfAllPOnlyAvarIdxSets);

        // Check condition for every sQ:
        outer:
        for (Variable x : this.getVars()) {
            Set<Integer> sQ = Sets.newHashSet(x.getIndices());
            sQ.retainAll(allAvarIndices); // Only avars!

            // Do the check only if there are >= 2 avars identified in q:
            if (sQ.size() >= 2) {
                //System.out.println("Checking if sQ identifies too much: " + sQ);
                for (Set<Integer> sP : setOfAllPOnlyAvarIdxSets) {
                    if (sP.containsAll(sQ)) {
                        //System.out.println("No, found matching sP:             " + sP);
                        if (!pIdxSet2VariableMap.containsKey(sP)) {
                            throw new CQException("[" + this.getClass().getSimpleName() 
                                    + "] Key not in map: " + sP);
                        }
                        continue outer;
                    }
                }
                // Did not find a matching sP for sQ!
                return true;
            } 
        }
        return false;
    }

    /**
     * Generate a random rCQ of given length.
     *
     * @param numberOfVars
     * @param sigma
     * @return
     * @throws de.unibremen.informatik.tdki.grind2.cq.CQException
     */
    public static RootedCQ
            getRandomRootedCQ(int numberOfVars, Signature sigma) throws CQException {
        if (sigma.getSetOfClasses().isEmpty()) {
            throw new CQException("[" + RootedCQ.class
                    .getSimpleName()
                    + "] Tried to construct RootedCQ with empty class signature!");

        }
        if (sigma.getSetOfRoles().isEmpty()) {
            throw new CQException("[" + RootedCQ.class
                    .getSimpleName()
                    + "] Tried to construct RootedCQ with empty role signature!");

        }
        if (numberOfVars == 0) {
            throw new CQException("[" + RootedCQ.class
                    .getSimpleName()
                    + "] Tried to construct RootedCQ with 0 variables!");
        }

        // Create all vars:
        List<Variable> listOfVars = new ArrayList<>();
        for (int i = 0; i < numberOfVars; i++) {
            listOfVars.add(new Variable());
        }

        // Randomly select some as avars, at least one:
        Set<Variable> setOfAvars = new HashSet<>();
        selectRandomAvars(setOfAvars, numberOfVars, listOfVars);

        // Put classes and roles into Lists:
        List<OWLClass> listOfClasses
                = new ArrayList<>(sigma.getSetOfClasses());
        List<OWLObjectProperty> listOfRoles
                = new ArrayList<>(sigma.getSetOfRoles());

        // Incidence matrix for classes and roles:
        int numberOfClasses = listOfClasses.size();
        int numberOfRoles = listOfRoles.size();
        int sizeOfCRArray = numberOfClasses + numberOfRoles * numberOfVars;
        boolean[][] classAndRoleArray = new boolean[sizeOfCRArray][numberOfVars];

        // Array with truth values for "variable already used". Defaults to false.
        boolean[] varUsed = new boolean[numberOfVars];

        // Connectedness algorithm by spantree:
        assureConnectedness(classAndRoleArray, numberOfVars, numberOfRoles, numberOfClasses);

        if (numberOfVars == 1) {
            assureAtLeastOneOccurrence(varUsed, sizeOfCRArray, numberOfClasses,
                    classAndRoleArray, numberOfRoles);
        }

        fillMatrixRandomly(sizeOfCRArray, numberOfVars, classAndRoleArray);

        //printMatrix(sizeOfCRArray, numberOfVars, classAndRoleArray);
        Set<Atom> setOfAtoms = new HashSet<>();
        // Fill class atoms:
        for (int line = 0; line < numberOfClasses; line++) {
            for (int varNr = 0; varNr < numberOfVars; varNr++) {
                if (classAndRoleArray[line][varNr]) {
                    Atom at = new ClassAtom(listOfClasses.get(line),
                            listOfVars.get(varNr));
                    setOfAtoms.add(at);
                }
            }
        }
        // Fill role atoms:
        for (int line = numberOfClasses; line < sizeOfCRArray; line++) {
            for (int varNr = 0; varNr < numberOfVars; varNr++) {
                if (classAndRoleArray[line][varNr]) {
                    Atom at = new RoleAtom(
                            listOfRoles.get((line - numberOfClasses) / numberOfVars),
                            listOfVars.get(varNr),
                            listOfVars.get((line - numberOfClasses) % numberOfVars));
                    setOfAtoms.add(at);
                }
            }
        }
        RootedCQ rCQ = new RootedCQ(setOfAvars, setOfAtoms);
        return rCQ;
    }

    private static void assureConnectedness(boolean[][] classAndRoleArray,
            int numberOfVars, int numberOfRoles, int numberOfClasses) {

        // Build random span tree to guarantee connectedness:
        // - x0 is connected; - find random neighbor; - iterate
        List<Integer> unconnectedVars = new ArrayList<>();
        for (int i = 1; i < numberOfVars; i++) {
            unconnectedVars.add(i);
        }
        int lookingForNeighbor = 0;
        Random rand = new Random();
        while (!unconnectedVars.isEmpty()) {
            // Get random neighbor:
            int neighborIndex = rand.nextInt(unconnectedVars.size());
            int neighbor = unconnectedVars.get(neighborIndex);
            unconnectedVars.remove(neighborIndex);
            // Decide role to use:
            int usedRole = rand.nextInt(numberOfRoles);
            // Decide direction of edge:
            if (Math.random() < 0.5) {
                // x->neighbor
                classAndRoleArray[numberOfClasses + usedRole * numberOfVars
                        + lookingForNeighbor][neighbor] = true;
            } else {
                // x<-neighbor
                classAndRoleArray[numberOfClasses + usedRole * numberOfVars
                        + neighbor][lookingForNeighbor] = true;
            }
            lookingForNeighbor = neighbor;
        }
    }

    private static void assureAtLeastOneOccurrence(boolean[] varUsed, int sizeOfCRArray,
            int numberOfClasses, boolean[][] classAndRoleArray, int numberOfRoles) {
        // Distribute one occurrence of variable 0:
        // Find a place where to use x_i:
        Random rand = new Random();
        int whereToUse = rand.nextInt(sizeOfCRArray);
        if (whereToUse < numberOfClasses) {
            // Ok, it is a class atom:
            classAndRoleArray[whereToUse][0] = true;
        } else {
            // It is a role atom:
            classAndRoleArray[whereToUse][0] = true;
        }
        varUsed[0] = true;
    }

    private static void selectRandomAvars(Set<Variable> setOfAvars,
            int numberOfVars, List<Variable> listOfVars) {

        Set<Integer> avarIndexSet = new HashSet<>();
        Random rand = new Random();
        // Add one variable randomly:
        avarIndexSet.add(rand.nextInt(numberOfVars));
        // Eventually add some more randomly:
        for (int varNr = 0; varNr < numberOfVars; varNr++) {
            if (Math.random() < CQ.getAVARPROB()) {
                avarIndexSet.add(varNr);
            }
        }
        for (Integer i : avarIndexSet) {
            setOfAvars.add(listOfVars.get(i));
        }
    }

    private static void fillMatrixRandomly(int sizeOfCRArray, int numberOfVars,
            boolean[][] classAndRoleArray) {

        // Now fill the rest of the matrix randomly:
        for (int line = 0; line < sizeOfCRArray; line++) {
            for (int varNr = 0; varNr < numberOfVars; varNr++) {
                if (Math.random() < CQ.getATOMPROB()) {
                    classAndRoleArray[line][varNr] = true;
                }
            }
        }
    }

    private static void printMatrix(int sizeOfCRArray, int numberOfVars,
            boolean[][] classAndRoleArray) {

        for (int line = 0; line < sizeOfCRArray; line++) {
            for (int varNr = 0; varNr < numberOfVars; varNr++) {
                if (classAndRoleArray[line][varNr]) {
                    System.out.print("1");
                } else {
                    System.out.print("0");
                }
            }
            System.out.println();
        }
    }

} // class RootedCQ
