/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.cq;

import de.unibremen.informatik.tdki.grind2.cq.components.Atom;
import de.unibremen.informatik.tdki.grind2.cq.components.ClassAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.RoleAtom;
import de.unibremen.informatik.tdki.grind2.cq.components.Variable;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

/**
 *
 * @author hansen
 */
public class Splitting {

    Partition partition;
    Map<Variable, Variable> nu;

    public Splitting(Partition partition, Map<Variable, Variable> nu) {
        this.partition = partition;
        this.nu = nu;
    }

    /**
     * This checks if the generated Splitting actually describes a valid match
     * into p. To check this, we need to see if all roles of R atoms under nu
     * are present in p.
     *
     * @param p
     * @return
     */
    public boolean atomsExistingIn(RootedCQ p) {
        Set<Atom> rAtomSet = this.partition.getrPart().getSetOfAtoms();
        Set<Atom> imageOfRUnderNu = new HashSet<>();
        
        for (Atom at : rAtomSet) {
            if (at instanceof ClassAtom) {
                ClassAtom cat = (ClassAtom) at;
                Variable x = cat.getVar();
                Atom imageOfCat = new ClassAtom(cat.getCls(), nu.get(x));
                imageOfRUnderNu.add(imageOfCat);
            }
            if (at instanceof RoleAtom) {
                RoleAtom rat = (RoleAtom) at;
                Variable x = rat.getXVar();
                Variable y = rat.getYVar();
                Atom imageOfRat = new RoleAtom(rat.getRole(), nu.get(x), nu.get(y));
                imageOfRUnderNu.add(imageOfRat);
            }
        }
        
        return p.getSetOfAtoms().containsAll(imageOfRUnderNu);
    }

    public Partition getPartition() {
        return partition;
    }

    public Map<Variable, Variable> getNu() {
        return nu;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 11 * hash + Objects.hashCode(this.partition);
        hash = 11 * hash + Objects.hashCode(this.nu);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Splitting other = (Splitting) obj;
        if (!Objects.equals(this.partition, other.partition)) {
            return false;
        }
        if (!Objects.equals(this.nu, other.nu)) {
            return false;
        }
        return true;
    }

    @Override
    public String toString() {
        return "Splitting{" + partition + ", nu=" + nu + '}';
    }

}
