/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.runnables;

import de.unibremen.informatik.tdki.grind.context.OriginalOntology;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.control.OntPlusQueryRunner;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Set;
import net.sourceforge.argparse4j.ArgumentParsers;
import net.sourceforge.argparse4j.impl.Arguments;
import net.sourceforge.argparse4j.inf.ArgumentParser;
import net.sourceforge.argparse4j.inf.ArgumentParserException;
import net.sourceforge.argparse4j.inf.FeatureControl;
import net.sourceforge.argparse4j.inf.MutuallyExclusiveGroup;
import net.sourceforge.argparse4j.inf.Namespace;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;

/**
 *
 * @author hansen
 */
public class MainRunnable {
    
    private static final int SECOND = 1000; // 1 s = 1000 ms
    private static final int MINUTE = 60_000;
    private static final int HOUR = 60 * 60_000;
    
    final static String VERSION = "2.1 (2017-05-30)";

    //Example:
    //   $ grind ./ontologies/envo.owl ./queries/envo.txt
    // 
    // Output will be rewritings of all queries in the file, in Datalog syntax
    // to stdout, e.g.
    //   Q(?0)  <-  owlENVO_00002043(?0)
    //   Q(?0)  <-  owlENVO_00002272(?0)
    //   Q(?0)  <-  owlENVO_00003043(?0)
    //
    public static void main(String[] args) throws OWLOntologyCreationException, FileNotFoundException, Exception {
        
        // Parsing of command line arguments is done using argparse4j. 
        // Argument parsing in java is a royal pain, let's hope this
        // library stays around for some time.
        ArgumentParser argParser = ArgumentParsers.newArgumentParser("grind")
                .description("Rewrite queries into non-recursive datalog with "
                        + "respect to the ontology. Result is written to stdout.");
        
        argParser.addArgument("-v", "--verbose")
                .dest("verbose")
                .action(Arguments.storeConst())
                .setConst(true)
                .setDefault(false)
                .help("dump internals to stdout");
        
        argParser.addArgument("ontology")
                .help("an OWL ontology, e.g. './ontologies/envo.owl'");
        
        MutuallyExclusiveGroup group = argParser.addMutuallyExclusiveGroup("type of query")
                .required(true);
        group.addArgument("-f", "--queryfile")
                .dest("queryfile")
                .help("query file, e.g. './queries/envo.txt'");
        group.addArgument("-q", "--query")
                .dest("query")
                //.action(Arguments.storeConst())
                //.setConst(true)
                .help("provide single query on command line; use quotation marks, e.g.\n-q \"Q(?0) <- Bacillus(?0)\"");
        
        argParser.addArgument("-t", "--timeout")
                .dest("timeout")
                .setDefault(30 * SECOND)
                .help("timeout [s] for single call of AQ algorithm. Defaults to 30 s.");
        
        argParser.epilog("Grind -- Gracefully Rewriting Into Non-recursive Datalog.\n"
                + "Version " + VERSION + ".\n"
                + "Copyright 2014, 2015 Peter Hansen and İnanç Seylan,\n"
                + "          2016, 2017 Peter Hansen.\n"
                + "University of Bremen, AG TdKI.\n"
                + "Homepage: http://www.informatik.uni-bremen.de/~hansen/grind\n"
                + "This software is released under GPL 3 licence.\n"
        //                + "\n"
        //                + "Supports the following types of axioms in an ontology:\n"
        //                + " * <SubClassOf>\n"
        //                + " * <EquivalentClasses>\n"
        //                + " * <SubObjectPropertyOf>\n"
        //                + " * <EquivalentObjectProperties>\n"
        //                + " * <ObjectPropertyDomain>\n"
        //                + "\n"
        //                + "and the following concept constructors:\n"
        //                + " * <ObjectIntersectionOf>\n"
        //                + " * <ObjectSomeValuesFrom>\n"
        );
        argParser.version(VERSION);
        
        Namespace ns = null;
        try {
            ns = argParser.parseArgs(args);
        } catch (ArgumentParserException e) {
            argParser.handleError(e);
            System.exit(-1);
        }
        
        CQContext.getINSTANCE().setVerbose(ns.getBoolean("verbose"));
        
        CQContext.getINSTANCE().setTimeout(Integer.parseInt(ns.getString("timeout")));
        
        OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
        String ontfileString = ns.getString("ontology");
        OWLOntology fromFileOnt = manager.loadOntologyFromOntologyDocument(new File(ns.getString("ontology")));
        OriginalOntology originalOnt = new OriginalOntology(fromFileOnt, manager);
        Set<OWLAxiom> axioms = originalOnt.getGCIsAsAxioms();
        
        OntPlusQueryRunner runner = new OntPlusQueryRunner(axioms, ontfileString);
        
        if ((ns.getString("queryfile") != null)) {
            CQContext.getINSTANCE().setQueryfile(true);
            runner.runQueryfile(ns.getString("queryfile"));
        } else {
            CQContext.getINSTANCE().setQueryfile(false);
            runner.runAdhocquery(ns.getString("query"));
        }
        
        System.exit(0);
    }
}
