/*
 * Copyright 2014, 2015 Peter Hansen and İnanç Seylan,
 *           2016, 2017 Peter Hansen.
 * 
 * This file is part of Grind v2.
 * 
 * Grind is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Grind is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Grind.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibremen.informatik.tdki.grind2.utilities;

import de.unibremen.informatik.tdki.grind.context.OriginalOntology;
import de.unibremen.informatik.tdki.grind2.context.CQContext;
import de.unibremen.informatik.tdki.grind2.cq.CQException;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogProgram;
import de.unibremen.informatik.tdki.grind2.datalog.DatalogRule;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogClassAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogQueryAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogRoleAtom;
import de.unibremen.informatik.tdki.grind2.datalog.components.DatalogVariable;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.OWLAxiom;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLObjectProperty;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLSubClassOfAxiom;
import org.semanticweb.owlapi.model.PrefixManager;

/**
 *
 * @author hansen
 */
public class OntologyUtils {

    public static OWLObjectProperty createRole(String name, PrefixManager pm) {
        return OWLManager.getOWLDataFactory().getOWLObjectProperty(":" + name, pm);
    }

    public static OWLClass createClass(String name, PrefixManager pm) {
        return OWLManager.getOWLDataFactory().getOWLClass(":" + name, pm);
    }

    public static OWLClassExpression getClsOrIntersection(Set<OWLClassExpression> setOfClasses) {
        if (setOfClasses.isEmpty()) {
            //            throw new CQException("[" + AvarReasoner.class.getSimpleName()
            //                    + "] tlcs is empty!");
            return OWLManager.getOWLDataFactory().getOWLThing();
        }
        OWLClassExpression resultConcept;
        if (setOfClasses.size() == 1) {
            resultConcept = setOfClasses.iterator().next();
        } else {
            resultConcept = OWLManager.getOWLDataFactory().getOWLObjectIntersectionOf(setOfClasses);
        }
        return resultConcept;
    }

    public static OWLOntology getOntologyPlusNewAxioms(OriginalOntology oldOnt, Set<OWLAxiom> newAxioms)
            throws CQException, OWLOntologyCreationException {
        Set<OWLAxiom> ontPlusNewAxioms = new HashSet<>();
        for (OWLSubClassOfAxiom ci : oldOnt.getGCIs()) {
            ontPlusNewAxioms.add((OWLAxiom) ci);
        }
        for (OWLAxiom ci : newAxioms) {
            ontPlusNewAxioms.add(ci);
        }
        OWLOntology optimizationOntology = OWLManager.createOWLOntologyManager().createOntology(ontPlusNewAxioms);
        return optimizationOntology;
    }

    public static DatalogProgram replaceNbyQReplaceAnnotationsAndAddCore(DatalogProgram program,
            Set<DatalogRoleAtom> coreDatalogRoleAtoms, Set<DatalogVariable> coreDatalogVariables) {

        // We need to find the highest var number used in the core. 
        // Fresh variables needed now will get a higher number!
        Set<Integer> coreVarIndices = new HashSet<>();
        for (DatalogVariable x : coreDatalogVariables) {
            coreVarIndices.add(x.getNumber());
        }
        // Can start assigning fresh variables with max+1:
        int freshVarCounter = Collections.max(coreVarIndices) + 1;

        Set<DatalogRule> setOfRules = new HashSet<>();

        // Iterate over Rules; replace  N  by  Q(?0, ?1, ?2)  and Annotations 
        // A^x[0], B^x[1]  by  A(?0), B(?1)
        outer:
        for (DatalogRule rule : program.getSetOfRules()) {

            // Change head if it's N:
            DatalogAtom head = rule.getHeadAtom();
            if (head instanceof DatalogClassAtom) {
                DatalogClassAtom dCat = (DatalogClassAtom) head;
                if (dCat.getConcept().equals(CQContext.getINSTANCE().getClsN())) {
                    head = new DatalogQueryAtom(coreDatalogVariables);
                }
            }

            Set<DatalogAtom> bodyAtoms = rule.getBody();
            // Omit body atom if it contains N: 
            for (DatalogAtom at : bodyAtoms) {
                if (at instanceof DatalogClassAtom) {
                    DatalogClassAtom dcat = (DatalogClassAtom) at;
                    if (dcat.getConcept().equals(CQContext.getINSTANCE().getClsN())) {
                        // Don't add this atom to new atoms!
                        continue outer;
                    }
                }
            }

            // Always add all core role atoms:
            Set<DatalogAtom> newBodyAtoms = new HashSet<>(coreDatalogRoleAtoms);

            for (DatalogAtom bodyAtom : bodyAtoms) {

                // ClassAtoms:
                if (bodyAtom instanceof DatalogClassAtom) {
                    DatalogClassAtom classBodyAtom = (DatalogClassAtom) bodyAtom;
                    OWLClass cls = classBodyAtom.getConcept();

                    // Need to delete annotation if there is one.
                    // A^x[2](?0)  becomes  A(?2)
                    if (CQContext.getINSTANCE().getAnnotatedClasses2VarsMap().containsKey(cls)) {
                        int varNumber = CQContext.getINSTANCE().getAnnotatedClasses2VarsMap().get(cls);
                        DatalogVariable x = new DatalogVariable(varNumber);
                        OWLClass classWithoutAnnotation = CQContext.getINSTANCE().getAnnotatedClasses2ClassMap().get(cls).asOWLClass();
                        DatalogClassAtom newAtomWithoutAnnotation = new DatalogClassAtom(classWithoutAnnotation, x);
                        newBodyAtoms.add(newAtomWithoutAnnotation);
                    } else {
                        newBodyAtoms.add(bodyAtom);
                    }

                } else if (bodyAtom instanceof DatalogRoleAtom) {
                    DatalogRoleAtom roleBodyAtom = (DatalogRoleAtom) bodyAtom;
                    OWLObjectProperty role = roleBodyAtom.getRole();

                    // Need to delete annotation if there is one. 
                    // r^x[2](?0, ?1)  becomes  r(?2, ?7)
                    // ?7 is fresh var! 
                    if (CQContext.getINSTANCE().getAnnotatedRoles2VarsMap().containsKey(role)) {
                        int varNumber = CQContext.getINSTANCE().getAnnotatedRoles2VarsMap().get(role);
                        DatalogVariable x = new DatalogVariable(varNumber);
                        OWLObjectProperty roleWithoutAnnotation = CQContext.getINSTANCE().getAnnotatedRoles2RoleMap().get(role);
                        // Fresh var:
                        DatalogVariable freshVar = new DatalogVariable(freshVarCounter++);
                        DatalogRoleAtom newAtomWithoutAnnotation = new DatalogRoleAtom(roleWithoutAnnotation, x, freshVar);
                        newBodyAtoms.add(newAtomWithoutAnnotation);
                    } else {
                        newBodyAtoms.add(bodyAtom);
                    }
                } else {
                    System.err.println("[" + OntologyUtils.class.getSimpleName()
                            + "] Datalog body atom that is neither Class nor Role atom.");
                    System.exit(-1);
                }
            }
            DatalogRule newRule = new DatalogRule(head, newBodyAtoms);
            setOfRules.add(newRule);
        }

        DatalogProgram resultProgram = new DatalogProgram(setOfRules);
        return resultProgram;
    }

}
